/******************************************************************************
Accord filter function for vapoursynth by V.C.Mohan
The aClip  folds out or b Clip folds In accordian style in single or twin sections
6 sep 2015 3 june 2021
Author V.C,Mohan


    Copyright (C) <2015-2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
	
********************************************************************************/ 
// for independant compilig to test remove comment brackets
/*
#include <stdlib.h>
#include "VapourSynth.h"
#include "VSHelper.h"
#include "math.h"
#include "Linez.cpp"

*/
//------------------------------------------------------------------
// shading functios for shading alternate folds of accordian
template <typename finc>
void shadeH( finc *dp,int dpitch,  int nfoldh, int wd, int dy, float shade);
template <typename finc>
void shadeW(finc *dp, int dpitch, int nfoldw, int ht, int dx,  float shade);
template <typename finc>
void drawHorizontalFoldLines(finc * dp,int dpitch,  int nfoldh, int wd, int dy, finc shade);
template <typename finc>
void drawVerticalFoldLines(finc * dp,int dpitch,  int nfoldw, int ht, int dx, finc shade);
//----------------------------------------------------------------------

template <typename finc>
void shadeW(finc *dp, int dpitch, int nfoldw, int ht, int dx, float shade)
{
	for(int f = 0; f < nfoldw ; f += 2)
	{
		for(int w = f * dx / nfoldw; w < (f + 1) * dx / nfoldw; w ++)
		{
			finc * d1p = dp; 

			for(int h = 0; h < ht; h ++)
			{				
				d1p[ w ] = (finc)(d1p[w] * shade)  ;

				d1p += dpitch;
			}
		}
	}
}
//-------------------------------------------------------------------------
template <typename finc>
void shadeH( finc *dp,int dpitch,  int nfoldh, int wd, int dy, float shade)
{
	for(int f = 0; f < nfoldh ; f += 2)		
	{
		for(int h = f * dy / nfoldh; h < (f + 1) * dy / nfoldh; h ++)
		{
			int hoffset = h * dpitch;

			for(int w = 0; w < wd; w ++)						
			{
				dp[hoffset + w]	= (finc)(dp[ hoffset + w] * shade) ;
			}
		}
	}
}

//-----------------------------------------------------------------------------------------
template <typename finc>
void drawHorizontalFoldLines(finc * dp,int dpitch,  int nfoldh, int wd, int dy, finc shade)
{
	for(int f = 0; f < nfoldh ; f += 2)		
	{
		int hoffset = ((f * dy) / nfoldh) * dpitch; 
		
		for(int w = 0;w < wd; w ++)						
		{
			dp[hoffset + w]	= shade ;
		}
	} 
}
//-------------------------------------------------------------------------------------
template <typename finc>
void drawVerticalFoldLines(finc *dp,int dpitch,  int nfoldw, int ht, int dx, finc shade)
{
	for(int f = 0; f < nfoldw ; f += 2)
	{
		int woffset = (f * dx) / nfoldw;

		finc * d1p = dp;
		
		for(int h = 0; h < ht; h ++)
		{				
			d1p[ woffset ] =  shade  ;

			d1p += dpitch;
		}
	}
}
//-------------------------------------------------------------------

typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap;
	bool vert;
	bool twin ;
	bool open ;

	int vtransStart;
	int vtransEnd;
	int numFrames;

} AccordData;
//---------------------------------------------------------
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC accordInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    AccordData *d = (AccordData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);
}
	
//-----------------------------------------------------------------
	// This is the main function that gets called when a frame should be produced. It will, in most cases, get
// called several times to produce one frame. This state is being kept track of by the value of
// activationReason. The first call to produce a certain frame n is always arInitial. In this state
// you should request all the input frames you need. Always do it in ascending order to play nice with the
// upstream filters.
// Once all frames are ready, the filter will be called with arAllFramesReady. It is now time to
// do the actual processing.
static const VSFrameRef *VS_CC accordGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
    AccordData *d = (AccordData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;
        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);


        // When creating a new frame for output it is VERY EXTREMELY SUPER IMPORTANT to
        // supply the "dominant" source frame to copy properties from. Frame props
        // are an essential part of the filter chain and you should NEVER break it.
    //    VSFrameRef *dst = vsapi->newVideoFrame(fi, width, height, srca, core);

		
		int nfoldw= (width/16) & 0xfffffffe; // number folds
		int nfoldh =(height/16) & 0xfffffffe;

		int dx = (((bn + 1) * nfoldw * 16)/(d->overlap + 2 ))  & 0xfffffff8;	// width of opening frame
		int dy = (((bn + 1) * nfoldh * 16)/(d->overlap + 2 ))  & 0xfffffff8;  // height of opening frame
		float shade;
		
	//	float  taperh, taperw;	// width or height of portion to be blackened to get fold effect
	//	float taper=3.0;						// trial & error arrived at value
				
		// too small a differance to process
		if (dy < 8 || dx < 8)
		{
			vsapi->freeFrame(srcb);
			return srca;
		}			

		else if (dy > height - 8 || dx > width - 8)
		{
			vsapi->freeFrame(srca);		
			return srcb;
		}
		const VSFrameRef * src, * base;
		// reset up to select the input frame for resizing and shading to mimic folding 
		if(d->open)
		{
			src = srca;
			base = srcb;
			dx = width - dx;							// width of visible screen
			dy = height - dy;							// height of visible screen
			shade = 1.0f - (0.25f * bn) / d->overlap;	// alternate fold sides to be darkened				
		}
		else
		{			
			src = srcb;
			base = srca;
			shade = 0.75f + 0.25f * bn / d->overlap;
		}

		if(d->twin)
		{
			dx= dx/2;
			dy= dy/2;
			nfoldh /= 2;
			nfoldw /= 2;
		}
			
	
		VSFrameRef * dst = vsapi->copyFrame(base, core);
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;
		int nplanes = fi->numPlanes;
		
		float by = (1.0f * height)/ dy;
		float bx = (1.0f * width) / dx;

		if(d->twin)
		{
			by /= 2;
			bx /= 2;
		}
		
		for(int plane = 0; plane < nplanes; plane ++)
		{
						
			int subw = plane > 0 && fi ->colorFamily == cmYUV? fi->subSamplingW : 0;
			int	subh = plane > 0 && fi ->colorFamily == cmYUV? fi->subSamplingH : 0;
			
			int d1y = dy >> subh;
			int d1x = dx >> subw;			

			const uint8_t *srcp = vsapi->getReadPtr(src, plane);
            int src_stride = vsapi->getStride(src, plane);
            uint8_t *dstp = vsapi->getWritePtr(dst, plane);
             // note that if a frame has the same dimensions and format, the stride is guaranteed to be the same. int dst_stride = src_stride would be fine too in this filter.
            // Since planes may be subsampled you have to query the height of them individually
            int bht = vsapi->getFrameHeight(src, plane);            
            int bwd = vsapi->getFrameWidth(src, plane);			
			int pitch = src_stride / kb;			

			if(! d->vert)			
			{			//Horizontal folding
				if(d->twin)
				{
					if (fi->sampleType == stInteger && nb == 8)
					{
						uint8_t black = 0;

						for(int w = 0; w < bwd; w ++)
						{
							// resize the frame to the new contracted size. Only vertical axis
							ResizeVLinePlanar( srcp + pitch * (bht/4) + w, 
												dstp	+ pitch * d1y/2 + w,
												d1y, pitch, 
												pitch, by);

							ResizeVLinePlanar( srcp + 3 * (bht/4) * pitch + w, 
												dstp 	+ pitch * (bht-1- d1y / 2) + w,
												d1y,  pitch, 
												pitch, by);
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;

						shadeH( dstp, pitch, nfoldh, bwd, d1y, shade);
						shadeH( dstp + (bht - 1 - d1y) * pitch, pitch,   nfoldh, bwd, d1y, shade);

						drawHorizontalFoldLines(dstp, pitch, nfoldh, bwd, d1y, black);
						drawHorizontalFoldLines(dstp + (bht - 1 - d1y) * pitch, pitch, nfoldh, bwd, d1y, black);
						
					}

					else if (fi->sampleType == stInteger && nb > 8)
					{ 
						const uint16_t * sp = (const uint16_t *) srcp;
						uint16_t * dp = (uint16_t *) dstp;
						uint16_t black = 0;

						for(int w = 0; w < bwd; w ++)
						{							
							// resize the frame to the new contracted size. Only vertical axis
							ResizeVLinePlanar( sp + pitch * (bht/4) + w, 
												dp	+ pitch * d1y/2 + w,
												d1y, pitch, 
												pitch, by);

							ResizeVLinePlanar( sp + 3 * (bht/4) * pitch + w, 
												dp 	+ pitch * (bht-1- d1y / 2) + w,
												d1y,  pitch, 
												pitch, by);
						}
					
				
					// shade alternate segments
						if (fi->colorFamily == cmYUV && plane > 0)
							continue;

						shadeH( dp, pitch,   nfoldh, bwd, d1y, shade);
						shadeH( dp + (bht - 1 - d1y) * pitch,  pitch,   nfoldh, bwd, d1y, shade);

						drawHorizontalFoldLines(dp, pitch, nfoldh, bwd, d1y, black);
						drawHorizontalFoldLines(dp + (bht - 1 - d1y) * pitch,  pitch, nfoldh, bwd, d1y, black);

					}
					else if (fi->sampleType == stFloat)
					{
						const float * sp = (const float *) srcp;
						float * dp = (float *) dstp;
						float black = 0.0; 

						for(int w = 0; w < bwd; w ++)
						{							
							// resize the frame to the new contracted size. Only vertical axis
							ResizeVLinePlanar( sp + pitch * (bht/4) + w, 
												dp	+ pitch * d1y/2 + w,
												d1y, pitch, 
												pitch, by);

							ResizeVLinePlanar( sp + 3 * (bht/4) * pitch + w, 
												dp 	+ pitch * (bht-1- d1y / 2) + w,
												d1y,  pitch, 
												pitch, by);
						}
					
				
					// shade alternate segments
						if (fi->colorFamily == cmYUV && plane > 0)
							continue;

						shadeH( dp, pitch,   nfoldh, bwd, d1y, shade);
						shadeH( dp + (bht - 1 - d1y) * pitch,  pitch,   nfoldh, bwd, d1y, shade);
						drawHorizontalFoldLines(dp, pitch, nfoldh, bwd, d1y, black);
						drawHorizontalFoldLines(dp + (bht - 1 - d1y) * pitch, pitch, nfoldh, bwd, d1y, black);

					}
				}
		
				
				else // if(!twin)
				{
					if (fi->sampleType == stInteger && nb == 8)
					{
						uint8_t black = 0; 

						for(int w = 0; w < bwd; w ++)
						{
							// resize the frame to the new contracted size. Only vertical axis
							ResizeVLinePlanar( srcp + pitch * (bht/2) + w, 
												dstp	+ pitch * (d1y/2) + w,
												d1y, pitch, 
												pitch, by);
						}
						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						shadeH( dstp, pitch, nfoldh, bwd, d1y, shade);
						drawHorizontalFoldLines(dstp, pitch, nfoldh, bwd, d1y, black);
					}

					else if (fi->sampleType == stInteger && nb > 8)
					{ 
						const uint16_t * sp = (const uint16_t *) srcp;
						uint16_t * dp = (uint16_t *) dstp;
						uint16_t black = 0;

						for(int w = 0; w < bwd; w ++)
						{							
							// resize the frame to the new contracted size. Only vertical axis
							ResizeVLinePlanar( sp + pitch * (bht/2) + w, 
												dp	+ pitch * d1y/2 + w,
												d1y, pitch, 
												pitch, by);
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						shadeH( dp, pitch, nfoldh, bwd, d1y, shade);
						drawHorizontalFoldLines(dp, pitch, nfoldh, bwd, d1y, black);
					}

					else if (fi->sampleType == stFloat)
					{ 
						const float * sp = (const float *) srcp;
						float * dp = (float *) dstp;
						float black = 0.0;

						for(int w = 0; w < bwd; w ++)
						{							
							// resize the frame to the new contracted size. Only vertical axis
							ResizeVLinePlanar( sp + pitch * (bht/2) + w, 
												dp	+ pitch * d1y/2 + w,
												d1y, pitch, 
												pitch, by);
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						shadeH( dp, pitch, nfoldh, bwd, d1y, shade);
						drawHorizontalFoldLines(dp, pitch, nfoldh, bwd, d1y, black);
					}
				}// not twin
			}// dir = hor
			

			else if(d->vert)
			{
				if(d->twin)
				{
					if (fi->sampleType == stInteger && nb == 8)
					{
						const uint8_t * sp =  srcp;
						uint8_t * dp =  dstp;
						uint8_t black = 0;
						
						for(int h = 0; h < bht; h ++)
						{

							ResizeHLinePlanar (sp + bwd / 4,
													dp + d1x / 2,
													d1x, bx);


							ResizeHLinePlanar (sp + 3 * bwd / 4,
													dp + (bwd- 1 - d1x / 2),
													d1x, bx );

							sp += pitch;
							dp += pitch;
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						dp =  dstp;
						shadeW( dp, pitch, nfoldw, bht, d1x, shade);
						shadeW( dp + bwd - 1 - d1x, pitch, nfoldw, bht, d1x, shade);

						drawVerticalFoldLines( dp, pitch, nfoldw, bht, d1x, black);
						drawVerticalFoldLines(dp + bwd - 1 - d1x, pitch, nfoldw, bht, d1x, black);
					}

					else if (fi->sampleType == stInteger && nb > 8)
					{ 
						const uint16_t * sp = (const uint16_t *) srcp;
						uint16_t * dp = (uint16_t *) dstp;
						uint16_t black = 0; 

						for(int h = 0; h < bht; h ++)
						{

							ResizeHLinePlanar (sp + bwd / 4,
													dp + d1x / 2,
													d1x, bx);


							ResizeHLinePlanar (sp + 3 * bwd / 4,
													dp + (bwd- 1 - d1x / 2),
													d1x, bx );

							sp += pitch;
							dp += pitch;
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						dp = (uint16_t *) dstp;
						shadeW( dp, pitch, nfoldw, bht, d1x, shade);
						shadeW( dp + bwd - 1 - d1x, pitch, nfoldw, bht, d1x, shade);
						drawVerticalFoldLines( dp, pitch, nfoldw, bht, d1x, black);
						drawVerticalFoldLines(dp + bwd - 1 - d1x, pitch, nfoldw, bht, d1x, black);
					}

					else if (fi->sampleType == stFloat)
					{ 
						const float * sp = (const float *) srcp;
						float * dp = (float *) dstp;
						float black = 0.0;

						for(int h = 0; h < bht; h ++)
						{

							ResizeHLinePlanar (sp + bwd / 4,
													dp + d1x / 2,
													d1x, bx);


							ResizeHLinePlanar (sp + 3 * bwd / 4,
													dp + (bwd- 1 - d1x / 2),
													d1x, bx );

							sp += pitch;
							dp += pitch;
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						dp = (float *) dstp;
						shadeW( dp, pitch, nfoldw, bht, d1x, shade);
						shadeW( dp + bwd - 1 - d1x, pitch, nfoldw, bht, d1x, shade);
						drawVerticalFoldLines( dp, pitch, nfoldw, bht, d1x, black);
						drawVerticalFoldLines(dp + bwd - 1 - d1x, pitch, nfoldw, bht, d1x, black);
					}
				}			
			
				else if(! d->twin)
				{
					if (fi->sampleType == stInteger && nb == 8)
					{
						const uint8_t * sp = srcp;
						uint8_t * dp = dstp;
						uint8_t black = 0;

						for(int h = 0; h < bht; h ++)
						{

							ResizeHLinePlanar(sp + bwd / 2,
												dp + d1x / 2,
												d1x, bx);
							sp += pitch;
							dp += pitch;
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						dp = dstp;
						shadeW( dp, pitch, nfoldw, bht, d1x, shade);
						drawVerticalFoldLines( dp, pitch, nfoldw, bht, d1x, black);
					
					}

					else if (fi->sampleType == stInteger && nb > 8)
					{ 
						const uint16_t * sp = (const uint16_t *) srcp;
						uint16_t * dp = (uint16_t *) dstp;
						uint16_t black = 0;

						for(int h = 0; h < bht; h ++)
						{

							ResizeHLinePlanar(sp  + bwd / 2,
												dp + d1x / 2,
												d1x, bx);
							sp += pitch;
							dp += pitch;
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						
						dp = (uint16_t *) dstp;

						shadeW( dp, pitch, nfoldw, bht, d1x, shade);
						drawVerticalFoldLines( dp, pitch, nfoldw, bht, d1x, black);
					
					}

					else if (fi->sampleType == stFloat)
					{ 
						const float * sp = (const float *) srcp;
						float * dp = (float *) dstp;
						float black = 0;
						for(int h = 0; h < bht; h ++)
						{

							ResizeHLinePlanar(sp + bwd / 2,
												dp + d1x / 2,
												d1x, bx);
							sp += pitch;
							dp += pitch;
						}

						if (fi->colorFamily == cmYUV && plane > 0)
							continue;
						dp = (float *) dstp;
						shadeW( dp, pitch, nfoldw, bht, d1x, shade);
						drawVerticalFoldLines( dp, pitch, nfoldw, bht, d1x, black);
					
					}
				} // not twin
			}	// vert
				
		
		}	// plane

		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);

		return dst;

	}
	return 0;
}
	
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC accordFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    AccordData *d = (AccordData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC accordCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    AccordData d;
    AccordData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Accord: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Accord: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }	
	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Accord: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;
//	d.vi.numFrames += d.numFrames;	// does not work



	temp = !!vsapi->propGetInt(in, "dir", 0, &err);
	if (err)
	{
		d.vert = true;
	}
	else if ( temp == 0)
		d.vert = false;
	else
		d.vert = true;
	
	temp = !!vsapi->propGetInt(in, "twin", 0, &err);
	if (err)
	{
		d.twin = true;
	}
	else if ( temp == 0)
		d.twin = false;
	else
		d.twin = true;

	temp = !!vsapi->propGetInt(in, "open", 0, &err);
	if (err)
	{
		d.open = true;
	}
	else if ( temp == 0)
		d.open = false;
	else
		d.open = true;

	 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (AccordData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Bubbles", accordInit, accordGetFrame, accordFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "accord", "VapourSynth Bubbles function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Accord", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;twin:int:opt;open:int:opt;", accordCreate, 0, plugin);
}
*/