
 /*  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *   
	The Bubbles Transition function for vapoursynth creates Bubbles from Right Frame fragments 
	and floats them on to Left Frame randomly and bursts the Bubbles on landing
	First 3 Aug 2015  3 june 2021
	Author V.C.Mohan

	Copyright (C) <2015-2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
 *  *  *  *  *  *  *    *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  */


//--------------------------------------------------------------------  
#define NMAXBUBBLES 16 
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap;
	int vtransStart;
	int vtransEnd;
	int numFrames;
	int stat;
	int  rad;
	int nBubbles;		
	int  *nthbbl;
		
} BubblesData;	
//---------------------------------------------------------
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC bubblesInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    BubblesData *d = (BubblesData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);
	
	d->rad = d->vi.width>d->vi.height?d->vi.width / (2 * NMAXBUBBLES):d->vi.height / (2 * NMAXBUBBLES);
	d->rad = d->rad < 8 ? 8: d->rad;
	d->nBubbles = (d->vi.width / (2 * d->rad)) * (d->vi.height / (2 * d->rad));

	d->nthbbl = new int[d->nBubbles + 1];
	//get non repeating number 0 to nbubbles - 1
	for(int i = 0; i < d->nBubbles; i ++ )
	{
		d->nthbbl[i] = rand() % d->nBubbles;
		for (int k = 0;k<i;k ++ )			
			if(d->nthbbl[k] == d->nthbbl[i])
					i -- ;
	
	}
	d->nthbbl[d->nBubbles] = d->nthbbl[0];	// for getting last movement correctly


}
//---------------------------------------------------------------------


//_____________________________________________________________________

	
 /*  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  */ 	
	static const VSFrameRef *VS_CC bubblesGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
    BubblesData *d = (BubblesData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
	
		int mobile = d->stat == 0? 1 : 0;		
		double mag = 4.0;
		
		// number of bubbles along horizontal and vertical ;
		int nxbbl = width / (2 * d->rad);
		int nybbl = height / (2 * d->rad);

		float bubblesPerFrame = d->nBubbles / ((7.0f * d->overlap) / 8.0f);

		int nburst = (int)(bn > d->overlap / 8 ? (bn - d->overlap / 8) * bubblesPerFrame : 0);	// Bubbled some time back, formed squares

			// number of bubbles present in any frame
		int nexpanding = (int)((bn < d->overlap / 8 ? bn : bn > (7 * d->overlap) / 8 ? d->overlap - bn : d->overlap / 8) * bubblesPerFrame);

		int maxbubbles = (int)((d->overlap / 8) * bubblesPerFrame);	// max number of bubbles alive in a frame

		
		int subh[] = {0, 0, 0};
		int subw[] = {0, 0, 0};

		if (fi ->colorFamily == cmYUV)
		{				
			subw[1] = fi->subSamplingW;
			subh[1] = fi->subSamplingH;
			subw[2] = fi->subSamplingW;
			subh[2] = fi->subSamplingH;
		}

		VSFrameRef * dst = vsapi->copyFrame(srca, core);

		const uint8_t * srcp[] = {NULL, NULL, NULL};
		uint8_t * dstp[] = {NULL, NULL, NULL};
		int stride[] = {0, 0, 0};

		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;
		// side equals diameter
		int xd = (2 * d->rad) ;
		int yd = (2 * d->rad) ;;

		for(int plane = 0; plane < fi->numPlanes; plane ++)
		{
			srcp[plane] = vsapi->getReadPtr(srcb, plane);
            stride[plane] = vsapi->getStride(srcb, plane);
            dstp[plane] = vsapi->getWritePtr(dst, plane);

		//	int ht = vsapi->getFrameHeight(srcb, plane);            
		//	int wd = vsapi->getFrameWidth(srcb, plane);
			
		//	int pitch = stride[plane] / kb;
			
			for(int i = 0;i<nburst;i ++ )
			{
				int xcoord = (d->nthbbl[i] % nxbbl) * xd;
				int ycoord =  (d->nthbbl[i] / nxbbl) * yd ; //  * pitch;
				int offset = (ycoord >> subh[plane]) * stride[plane] + (xcoord >> subw[plane]) * kb;

				vs_bitblt(dstp[plane] + offset, stride[plane], srcp[plane] + offset, stride[plane],
						(xd >> subw[plane]) * kb, yd >> subh[plane]);
			}
		}

		for(int i = 0; i < nexpanding; i  ++ )
		{
				// bubbles while forming grow and move to destination
						//destination and source coordinates
						
			int xcoord0 = d->rad + (d->nthbbl[nburst + i] % nxbbl) * 2 * d->rad;

			int ycoord0 = d->rad + (d->nthbbl[nburst + i] / nxbbl) * 2 * d->rad;

						// start appearance coordinates
						// mobile = 0 if static opted otherwise 1
			int xcoord1 = d->rad + (d->nthbbl[nburst + i + mobile] % nxbbl) * 2 * d->rad;

			int ycoord1 = d->rad + (d->nthbbl[nburst + i + mobile] / nxbbl) * 2 * d->rad;
						// current coordinates
			int xcoord = xcoord0 - ((xcoord0 - xcoord1) * i) / maxbubbles;

			int ycoord = ycoord0 - ((ycoord0 - ycoord1) * i) / maxbubbles;

			int radius = ((maxbubbles - i) * d->rad) / maxbubbles;

			if(bn < d->overlap / 8)
			{
				radius = (bn * radius) / (d->overlap / 8);

				xcoord = xcoord1 + ((xcoord0 - xcoord1) * bn) / (d->overlap / 8);

				ycoord = ycoord1 + ((ycoord0 - ycoord1) * bn) / (d->overlap / 8);
			}
				
			
			if (fi->sampleType == stInteger && fi->bitsPerSample == 8)
			{			
				const uint8_t *sp = srcp[0];
				int spitch = stride[0] / kb;
				uint8_t *dp = dstp[0];

				const uint8_t *s1p = srcp[1];
				int s1pitch = stride[1] / kb;
				uint8_t *d1p = dstp[1];
				const uint8_t *s2p = srcp[2];
				uint8_t *d2p = dstp[2];
				if( fi->numPlanes == 1)
				{
					vsDiscoPlane(sp,  dp, radius, 
								xcoord, ycoord, height, width,
								xcoord1, ycoord1, height, width, 
								spitch, spitch,   mag ,  true,  0);
				}

				else
				{
					vsDiscoYUVPlanes(sp, dp, radius,
									xcoord, ycoord, height, width,
									xcoord1, ycoord1, height, width,
									spitch, spitch,
									s1p, s2p, s1pitch,  
									d1p, d2p, s1pitch,
									subh[1], subw[1],
									mag ,  true, 0  );
				}
			}
					
			else if (fi->sampleType == stInteger && nb > 8)
			{
				const uint16_t *sp = (const uint16_t *)srcp[0];
				int spitch = stride[0] / kb;
				uint16_t *dp = (uint16_t *)dstp[0];

				const uint16_t *s1p = (const uint16_t *)srcp[1];
				int s1pitch = stride[1] / kb;
				uint16_t *d1p = (uint16_t *)dstp[1];
				const uint16_t *s2p = (const uint16_t *)srcp[2];
				uint16_t *d2p = (uint16_t *)dstp[2];

				if( fi->numPlanes == 1)
				{
					vsDiscoPlane(sp,  dp, radius, 
								xcoord, ycoord, height, width,
								xcoord1, ycoord1, height, width,
								spitch, spitch,   mag ,  true,  0);
				}

				else
				{
					vsDiscoYUVPlanes(sp, dp, radius,
									xcoord, ycoord, height, width,
									xcoord1, ycoord1, height, width,
									spitch, spitch,
									s1p, s2p, s1pitch,  
									d1p, d2p, s1pitch,
									subh[1], subw[1],
									mag ,  true, 0  );
				}
			}
		

			else if (fi->sampleType == stFloat)
			{
				const float *sp = (const float *)srcp[0];
				int spitch = stride[0] / kb;
				float *dp = (float *)dstp[0];

				const float *s1p = (const float *)srcp[1];
				int s1pitch = stride[1] / kb;
				float *d1p = (float *)dstp[1];
				const float *s2p = (const float *)srcp[2];
				float *d2p = (float *)dstp[2];

				if( fi->numPlanes == 1)
				{
					vsDiscoPlane(sp,  dp, radius, 
								xcoord, ycoord, height, width,
								xcoord1, ycoord1, height, width,
								spitch, spitch,   mag ,  true,  0);
				}

				else
				{
					vsDiscoYUVPlanes(sp, dp, radius,
									xcoord, ycoord, height, width,
									xcoord1, ycoord1, height, width,
									spitch, spitch,
									s1p, s2p, s1pitch,  
									d1p, d2p, s1pitch,
									subh[1], subw[1],
									mag ,  true, 0  );
				}
			}
		} 
		
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}

	return 0;
}

//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC bubblesFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    BubblesData *d = (BubblesData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	delete []d->nthbbl;
    free(d);
}
//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC bubblesCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    BubblesData d;
    BubblesData *data;
    int err;
	//int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Bubbles: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Bubbles: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Bubbles: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;
//	char ms[256];
//	sprintf(ms, "start = %d, end = %d, nf = %d, over = %d, anf = %d, bnf = %d",d.vtransStart,d.vtransEnd,d.numFrames,d.overlap,d.vi.numFrames,d.bvi.numFrames);
//	vsapi->setError(out, ms);
//        vsapi->freeNode(d.node[0]);
//		vsapi->freeNode(d.node[1]);
//        return;
	d.stat = !!vsapi->propGetInt(in, "static", 0, &err);
	if( err)
		d.stat = 0;

	if( d.stat < 0 || d.stat > 1)
	{
		vsapi->setError(out, "Bubbles: static can be either 0 or 1 ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (BubblesData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Bubbles", bubblesInit, bubblesGetFrame, bubblesFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "bubbles", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Bubbles", "clip:clip;clipb:clip;overlap:float:opt;static:int:opt;", bubblesCreate, 0, plugin);
}
*/
 