/******************************************************************************
TransFlipPage filter function for vapoursynth by V.C.Mohan
The B Clip  comes into view when the A Clip is flipped as we turn a page
 of book in the leftection specified.
 The clips should have identical lengths, height, width and same pixel type 

 5 Aug 2015  17 july 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
************************************************************************************************/ 

 template <typename finc>
void HFlipPageTaper( finc * dp, int pitch, int ht,  int wd, int ytaper, finc val, bool invert);
 template <typename finc>
void VFlipPageTaper( finc * dp, int pitch, int ht, int wd,  int xtaper, finc val, bool invert);
 //--------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
template <typename finc>
void HFlipPageTaper( finc * dp, int pitch, int ht,  int wd, int ytaper, finc val, bool invert)
{
	if (invert)
	 {
		for ( int w = 0; w < wd; w ++)
		{
			int taper = (ytaper * (wd - w)) / wd;
							 
							 
			for ( int h = 0; h < taper; h ++)
			{
				dp[ h * pitch + w] = val;				//  top 
				dp[( ht - 1 - h) * pitch + w] = val;		// bot
				
			}
				
		}
	}
	else
	{
		for ( int w = 0; w < wd; w ++)
		{
			int taper = (ytaper *  w) / wd;
							 
			for ( int h = 0; h < taper; h ++)
			{
				dp[ h * pitch + w] = val;				// left top half
				dp[ (ht - 1 - h) * pitch + w] = val;		// right top half
				
			}
				
		}
	}
}
//-----------------------------------------------------------------------------------------------
 template <typename finc>
void VFlipPageTaper( finc * dp, int pitch, int ht, int wd, int xtaper, finc val, bool invert)
 {
	 if (invert)
	 {
		 for ( int h = 0; h < ht; h ++)
		{
			int taper = (xtaper * (ht - h)) / ht;
							 
			for ( int w = 0; w < taper; w ++)
			{
				dp[  w] = val;				// left top half
				dp[  wd - 1 - w] = val;		// right top half
			}
				dp += pitch;
		}
	 }
	 else
	 {
	 	// fill val in taper part
		for ( int h = 0; h < ht; h ++)
		{
			int taper = ( xtaper * h) / ht;
							 
			for ( int w = 0; w < taper; w ++)
			{
				dp[  w] = val;				// left top half
				dp[  wd - 1 - w] = val;		// right top half
			}
				dp += pitch;
		}
	 }
 }
//------------------------------------------------------------
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	bool left;
	int vtransStart;
	int vtransEnd;
	int numFrames;
		
}FlipPageData;

//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC flipPageInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    FlipPageData *d = (FlipPageData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

	
}
	
/***************************************************************/	
static const VSFrameRef *VS_CC flipPageGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     FlipPageData *d = (FlipPageData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		int deltay =((bn + 1) * height)/((d->overlap + 1));
		
		int deltax = ((bn + 1) * width)/((d->overlap + 1));
		
		int taperx = ((bn + 1) * (width /20) )/((d->overlap + 1));
		int tapery = ((bn + 1) * (height /20) )/((d->overlap + 1));

		if( (d->left  && deltax <= 4) || (! d->left 
			&& deltay <= 4) )
		{
			vsapi->freeFrame(srcb);
			return srca;
		}

		bool flip;

		if(bn > d->overlap / 2)
		{
					// flipped half way. So will start see the right frame
			deltay -=  height/2;
			deltax -=  width/2;
			flip = true;
			taperx = width / 20 - taperx ;
			tapery = height / 20 - tapery ;

			if ( ( d-> left  && deltay > height / 2 - 4) || ( !d->left  && deltax > width /2 - 4) )
			{
				vsapi->freeFrame(srca);
				return (srcb);
			}

		}
		else
			flip = false;

		deltax &= 0xfffffffc;
		deltay &= 0xfffffffc;
	
		int subh[] = {0, 0, 0, 0};
		int subw[] = {0, 0, 0, 0};

		if (fi ->colorFamily == cmYUV)
		{				
			subw[1] = fi->subSamplingW;
			subh[1] = fi->subSamplingH;
			subw[2] = fi->subSamplingW;
			subh[2] = fi->subSamplingH;
		}

		VSFrameRef * dst = vsapi->newVideoFrame(fi, width, height, srca, core);
		
		int nbytes = fi->bytesPerSample;
		int nbits = fi->bitsPerSample;

		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcpa = vsapi->getReadPtr(srca, plane);
			const uint8_t * srcpb = vsapi->getReadPtr(srcb, plane);
			int stride = vsapi->getStride(srca, plane);
			uint8_t *dstp = vsapi->getWritePtr(dst, plane);
			int ht = vsapi->getFrameHeight(srca, plane);
			int wd = vsapi->getFrameWidth(srca, plane);
			int pitch = stride / nbytes;
			int dy = deltay >> subh[plane];
			int dx = deltax >> subw[plane];
			int tapx = taperx >> subw[plane];
			int tapy = tapery >> subh[plane];

			if(! d->left )	// up			
			{			
				//B Frame turns upward covering out A Frame
			
				if(!(flip))
				{
				
					// copy A frame upper half
					vs_bitblt(dstp, stride, srcpa, stride, wd * nbytes, ht/2 );
				// copy bottom revealed part of B frame
					vs_bitblt(dstp + (ht - 1 - dy) * stride, stride, srcpb +(ht - 1 - dy) * stride, stride, wd * nbytes, dy); 

					if( fi->sampleType == stInteger && nbytes == 1)
					{
					
						uint8_t grey = plane != 0 && fi->colorFamily == cmYUV ?1 << (nbits - 1) : 0;

						for(int w = 0; w < wd; w ++)
						{
							ResizeVLinePlanar(srcpa + (3 * ht / 4) * pitch + w,
											dstp + (3 * ht / 4 - dy / 2) * pitch + w,
											ht / 2 - dy, pitch, pitch,
											ht / (ht - 2.0 * dy) );
						
						}

						VFlipPageTaper(dstp, pitch, ht / 2, wd, tapx, grey, false);
						// blacken raised taper part in the lower half of frame
						VFlipPageTaper(dstp + (ht / 2 ) * pitch, pitch, ht / 2 - dy, wd, tapx, grey, true);
						// blacken lower revealed taper part of frame
						VFlipPageTaper(dstp + (ht - 1 - dy) * pitch, pitch,  dy, wd, tapx, grey, true);

						drawHLinePlanar(dstp + ht / 2 * pitch, wd, grey);

					} 

					else if( fi->sampleType == stInteger && nbytes == 2)
					{
						const uint16_t * srpa = (const uint16_t * )srcpa;
						const uint16_t * srpb = (const uint16_t * ) srcpb;
						
						uint16_t * dp = (uint16_t * )dstp;
						uint16_t grey = plane != 0 && fi->colorFamily == cmYUV ?1 << (nbits - 1) : 0;
						
						
					// resize vertical of lower half of srca

						for(int w = 0; w < wd; w ++)
						{
							ResizeVLinePlanar(srpa + (3 * ht / 4) * pitch + w,
											dp + (3 * ht / 4 - dy / 2) * pitch + w,
											ht / 2 - dy, pitch, pitch,
											(ht - 1.0) / (ht - 2 * dy) );
						
						}

						VFlipPageTaper(dp, pitch, ht / 2, wd, tapx, grey, false);
						// blacken raised taper part in the lower half of frame
						VFlipPageTaper(dp + (ht / 2 ) * pitch, pitch, ht / 2 - dy, wd, tapx, grey, true);
						// blacken lower revealed taper part of frame
						VFlipPageTaper(dp + (ht - 1 - dy) * pitch, pitch,  dy, wd, tapx, grey, true);

						drawHLinePlanar(dp + ht / 2 * pitch, wd, grey);
					}
					else if( fi->sampleType == stFloat)
					{
						const float * srpa = (const float * )srcpa;
						const float * srpb = (const float * ) srcpb;
						
						float * dp = (float * )dstp;

						float grey = 0;
						
						
					// resize vertical of lower half of srca

						for(int w = 0; w < wd; w ++)
						{
							ResizeVLinePlanar(srpa + (3 * ht / 4) * pitch + w,
											dp + (3 * ht / 4 - dy / 2) * pitch + w,
											ht / 2 - dy, pitch, pitch,
											(ht - 1.0) / (ht - 2 * dy) );
						
						}

						VFlipPageTaper(dp, pitch, ht / 2, wd, tapx, grey, false);
						// blacken raised taper part in the lower half of frame
						VFlipPageTaper(dp + (ht / 2 ) * pitch, pitch, ht / 2 - dy, wd, tapx, grey, true);
						// blacken lower revealed taper part of frame
						VFlipPageTaper(dp + (ht - 1 - dy) * pitch, pitch,  dy, wd, tapx, grey, true);

						drawHLinePlanar(dp + ht / 2 * pitch, wd, grey);
					}					
				}												
			
				else if(flip)
				{
					//Second half when leaf becomes vertical and A Frame bottom half nolonger visible
					// B frame bottom half seen fully. A frame top part seen partly
					// B frame top part seen partly
					
					// copy B frame lower half -1, + 1 is to get central line
					vs_bitblt(dstp + (ht/ 2 ) * stride, stride, srcpb  + (ht/ 2 ) * stride, stride,  wd * nbytes, ht/2 );
					// copy top visible part of A frame
					vs_bitblt(dstp , stride, srcpa, stride, wd * nbytes, ht / 2 - dy ); 
				
					if( fi->sampleType == stInteger && nbytes == 1)
					{
					
						uint8_t grey = plane != 0 && fi->colorFamily == cmYUV ?1 << (nbits - 1) : 0;

						
						for(int w = 0 ; w < wd; w ++)
						{
							ResizeVLinePlanar(srcpb + ht/4 *  pitch + w,
											dstp + ( ht / 2 - dy / 2) * pitch + w,
											dy, pitch, pitch,
											ht / ( 2.0 * dy) );
						
						}
						// blacken upper part taper of still visible A frame
						VFlipPageTaper(dstp, pitch, ht / 2 - dy, wd, tapx, grey, false);
						// blacken raised taper part in the upper half of frame
						VFlipPageTaper(dstp + (ht / 2 - dy ) * pitch, pitch, dy, wd, tapx, grey, false);
						// blacken lower revealed taper part of frame
						VFlipPageTaper(dstp + (ht / 2) * pitch, pitch,  ht / 2, wd, tapx, grey, true);
						drawHLinePlanar(dstp + ht / 2 * pitch, wd, grey);
					}

					else if( fi->sampleType == stInteger && nbytes == 2)
					{
						const uint16_t * srpa = (const uint16_t * )srcpa;
						const uint16_t * srpb = (const uint16_t * ) srcpb;
						
						uint16_t * dp = (uint16_t * )dstp;
						uint16_t grey = plane != 0 && fi->colorFamily == cmYUV ?1 << (nbits - 1) : 0;
						
						for(int w = 0 ; w < wd; w ++)
						{
							ResizeVLinePlanar(srpb + ht/4 *  pitch + w,
											dp + ( ht / 2 - dy / 2) * pitch + w,
											dy, pitch, pitch,
											ht / ( 2.0 * dy) );
						
						}
						// blacken upper part taper of still visible A frame
						VFlipPageTaper(dp, pitch, ht / 2 - dy, wd, tapx, grey, false);
						// blacken raised taper part in the upper half of frame
						VFlipPageTaper(dp + (ht / 2 - dy ) * pitch, pitch, dy, wd, tapx, grey, false);
						// blacken lower revealed taper part of frame
						VFlipPageTaper(dp + (ht / 2) * pitch, pitch,  ht / 2, wd, tapx, grey, true);

						drawHLinePlanar(dp + ht / 2 * pitch, wd, grey);
					}

					else if( fi->sampleType == stFloat)
					{
						const float * srpa = (const float * )srcpa;
						const float * srpb = (const float * ) srcpb;
						
						float * dp = (float * )dstp;
						float grey = 0;
						
						for(int w = 0 ; w < wd; w ++)
						{
							ResizeVLinePlanar(srpb + ht/4 *  pitch + w,
											dp + ( ht / 2 - dy / 2) * pitch + w,
											dy, pitch, pitch,
											ht / ( 2.0 * dy) );
						
						}
						// blacken upper part taper of still visible A frame
						VFlipPageTaper(dp, pitch, ht / 2 - dy, wd, tapx, grey, false);
						// blacken raised taper part in the upper half of frame
						VFlipPageTaper(dp + (ht / 2 - dy ) * pitch, pitch, dy, wd, tapx, grey, false);
						// blacken lower revealed taper part of frame
						VFlipPageTaper(dp + (ht / 2) * pitch, pitch,  ht / 2, wd, tapx, grey, true);

						drawHLinePlanar(dp + ht / 2 * pitch, wd, grey);
					}				
				}

			}

			else // if( d->left  )	// left			
			{			
				//A Frame flips left  uncovering B Frame
			
				if(!(flip))
				{
				
					// copy A frame left half
					vs_bitblt(dstp, stride, srcpa, stride, wd /2 * nbytes, ht );
				// copy right revealed part of B frame
					vs_bitblt(dstp + (wd - 1 - dx) * nbytes, stride, srcpb +(wd - 1 - dx) * nbytes , stride, dx * nbytes, ht); 

					if( fi->sampleType == stInteger && nbytes == 1)
					{
					
						uint8_t grey = plane != 0 && fi->colorFamily == cmYUV ?1 << (nbits - 1) : 0;
						uint8_t * dp = dstp;
						const uint8_t * spa = srcpa;
						for(int h = 0; h < ht; h ++)
						{
							ResizeHLinePlanar(spa + (3 * wd / 4)  ,
											dp + (3 * wd / 4 - dx / 2),
											wd / 2 - dx,
											wd / (wd - 2.0 * dx) );
							spa += pitch;
							dp += pitch;
						
						}
						// blacken left half taper part
						HFlipPageTaper(dstp, pitch, ht, wd / 2, tapy, grey, false);
						// blacken raised taper part in the right half of frame
						HFlipPageTaper(dstp + wd / 2, pitch, ht, wd / 2 - dx, tapy, grey, true);
						// blacken right revealed taper part of frame
						HFlipPageTaper(dstp + (wd - 1 - dx) , pitch, ht, dx, tapy, grey, true);
						drawVLinePlanar(dstp + wd / 2 , pitch, ht, grey);

					} 

					else if( fi->sampleType == stInteger && nbytes == 2)
					{
						const uint16_t * srpa = (const uint16_t * )srcpa;
						const uint16_t * srpb = (const uint16_t * ) srcpb;
						
						uint16_t * dp = (uint16_t * )dstp;
						uint16_t grey = plane != 0 && fi->colorFamily == cmYUV ?1 << (nbits - 1) : 0;
						uint16_t * d1p = dp;
						const uint16_t * spa = srpa;

						for(int h = 0; h < ht; h ++)
						{
							ResizeHLinePlanar(spa + (3 * wd / 4)  ,
											d1p + (3 * wd / 4 - dx / 2),
											wd / 2 - dx,
											wd / (wd - 2.0 * dx) );
							spa += pitch;
							d1p += pitch;
						
						}
						// blacken left half taper part
						HFlipPageTaper(dp, pitch, ht, wd / 2, tapy, grey, false);
						// blacken raised taper part in the right half of frame
						HFlipPageTaper(dp + wd / 2, pitch, ht, wd / 2 - dx, tapy, grey, true);
						// blacken right revealed taper part of frame
						HFlipPageTaper(dp + (wd - 1 - dx) , pitch, ht, dx, tapy, grey, true);

						drawVLinePlanar(dp + wd / 2 , pitch, ht, grey);
					}
					else if( fi->sampleType == stFloat)
					{
						const float * srpa = (const float * )srcpa;
						const float * srpb = (const float * ) srcpb;						
						float * dp = (float * )dstp;
						float grey = 0;
						float * d1p = dp;
						const float * spa = srpa;

						for(int h = 0; h < ht; h ++)
						{
							ResizeHLinePlanar(spa + (3 * wd / 4)  ,
											d1p + (3 * wd / 4 - dx / 2),
											wd / 2 - dx,
											wd / (wd - 2.0 * dx) );
							spa += pitch;
							d1p += pitch;
						
						}
						// blacken left half taper part
						HFlipPageTaper(dp, pitch, ht, wd / 2, tapy, grey, false);
						// blacken raised taper part in the right half of frame
						HFlipPageTaper(dp + wd / 2, pitch, ht, wd / 2 - dx, tapy, grey, true);
						// blacken right revealed taper part of frame
						HFlipPageTaper(dp + (wd - 1 - dx) , pitch, ht, dx, tapy, grey, true);
						drawVLinePlanar(dp + wd / 2 , pitch, ht, grey);

					}					
				}												
			
				if(flip)
				{
					//Second half when leaf becomes vertical and A Frame right half nolonger visible
					// B frame right half seen fully. A frame left part seen partly
					// B frame left part seen partly
					
					// copy B frame right half 
					vs_bitblt(dstp + (wd/ 2 ) * nbytes, stride, srcpb  + (wd/ 2 ) * nbytes, stride,  wd / 2 * nbytes, ht );
					// copy left visible part of A frame
					vs_bitblt(dstp , stride, srcpa, stride, (wd / 2 - dx) * nbytes, ht  ); 
				
					if( fi->sampleType == stInteger && nbytes == 1)
					{
					
						uint8_t grey = plane != 0 && fi->colorFamily == cmYUV ?1 << (nbits - 1) : 0;
						uint8_t * d1p = dstp;
						const uint8_t * spb = srcpb;
						
						for(int h = 0 ; h < ht; h ++)
						{
							ResizeHLinePlanar(spb + wd/4 ,
											d1p + ( wd / 2 - dx / 2),
											dx, wd / ( 2.0 * dx) );

							spb += pitch;
							d1p += pitch;
						
						}
						// blacken left part taper of still visible A frame
						HFlipPageTaper(dstp, pitch, ht, wd / 2 - dx , tapy, grey, false);
						// blacken raised taper part in the left half of frame
						HFlipPageTaper(dstp + (wd / 2 - dx ), pitch, ht, dx, tapy, grey, false);
						// blacken right revealed taper part of frame
						HFlipPageTaper(dstp + (wd / 2), pitch,  ht, wd / 2, tapy, grey, true);
						drawVLinePlanar(dstp + wd / 2 , pitch, ht, grey);
					}

					else if( fi->sampleType == stInteger && nbytes == 2)
					{
						const uint16_t * srpa = (const uint16_t * )srcpa;
						const uint16_t * spb = (const uint16_t * ) srcpb;
						uint16_t * dp = (uint16_t * )dstp;
						uint16_t * d1p = (uint16_t * )dstp;
						uint16_t grey = plane != 0 && fi->colorFamily == cmYUV ?1 << (nbits - 1) : 0;
						
						for(int h = 0 ; h < ht; h ++)
						{
							ResizeHLinePlanar(spb + wd/4 ,
											d1p + ( wd / 2 - dx / 2),
											dx, wd / ( 2.0 * dx) );

							spb += pitch;
							d1p += pitch;
						
						}
						// blacken left part taper of still visible A frame
						HFlipPageTaper(dp, pitch, ht, wd / 2 - dx , tapy, grey, false);
						// blacken raised taper part in the left half of frame
						HFlipPageTaper(dp + (wd / 2 - dx ), pitch, ht, dx, tapy, grey, false);
						// blacken right revealed taper part of frame
						HFlipPageTaper(dp + (wd / 2), pitch,  ht, wd / 2, tapy, grey, true);
						drawVLinePlanar(dp + wd / 2 , pitch, ht, grey);
					}

					else if( fi->sampleType == stFloat)
					{
						const float * srpa = (const float * )srcpa;
						const float * srpb = (const float * ) srcpb;
						float * dp = (float * )dstp;
						float * d1p = (float * )dstp;
						const float * spb = srpb;
						float grey = 0;
						
						for(int h = 0 ; h < ht; h ++)
						{
							ResizeHLinePlanar(spb + wd/4 ,
											d1p + ( wd / 2 - dx / 2),
											dx, wd / ( 2.0 * dx) );

							spb += pitch;
							d1p += pitch;
						
						}
						// blacken left part taper of still visible A frame
						HFlipPageTaper(dp, pitch, ht, wd / 2 - dx , tapy, grey, false);
						// blacken raised taper part in the left half of frame
						HFlipPageTaper(dp + (wd / 2 - dx ), pitch, ht, dx, tapy, grey, false);
						// blacken right revealed taper part of frame
						HFlipPageTaper(dp + (wd / 2), pitch,  ht, wd / 2, tapy, grey, true);
						drawVLinePlanar(dp + wd / 2 , pitch, ht, grey);
					}				
				}

			}
		}// plane

			
			
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;
}

//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC flipPageFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    FlipPageData *d = (FlipPageData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC flipPageCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    FlipPageData d;
    FlipPageData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "FlipPage: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "FlipPage: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)(vsapi->propGetFloat(in, "overlap", 0, &err));
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "FlipPage: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	temp = !!vsapi->propGetInt(in, "left", 0, &err);
	if( err)
		d.left = true;

	if( temp == 0)
	{
		d.left = false;
	}
	else
		d.left = true;

	
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (FlipPageData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "FlipPage", flipPageInit, flipPageGetFrame, flipPageFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "flipPage", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("FlipPage", "clip:clip;clipb:clip;overlap:float:opt;left:int:opt;", flipPageCreate, 0, plugin);
}
*/
