/******************************************************************************
Trans.Funnel  plugin function for vapoursynth by V.C.Mohan
The A Clip  funnels out  in the direction specified,unmasking a stationary  B Clip .
 The clips should have identical lengths, height, width and same pixel type 

 5 Aug 2015  3 june 2021  17 july 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
********************************************************************************/  

//------------------------------------------------------------
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int dir;
	int vtransStart;
	int vtransEnd;
	int numFrames;
		
}FunnelData;
//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC funnelInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    FunnelData *d = (FunnelData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

	
}
/***************************************************************/	
static const VSFrameRef *VS_CC funnelGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     FunnelData *d = (FunnelData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);

		float deltay = 2.0f * ((bn + 1.0f) * height)/((d->overlap + 1));		
		float deltax = 2.0f * ((bn + 1.0f) * width)/((d->overlap + 1));

		if( ( (d->dir == 1 || d->dir == 3)  && deltax <= 4) || ( (d->dir == 2 || d->dir == 4) && deltay <= 4) )
		{
			vsapi->freeFrame(srcb);
			return srca;
		}
		
		if(2 * bn >= d->overlap)
		{
			deltax = (2.0f * (d->overlap - bn ) * width)/(d->overlap + 1);
			deltay = (2.0f * (d->overlap - bn) * height)/(d->overlap + 1);

		}

		int dropy = (height * (2 * bn - d->overlap)) / d->overlap;
		int dropx = (width * (2 * bn - d->overlap)) / d->overlap;

		VSFrameRef * dst = vsapi->copyFrame(srcb, core);

		int subh[] = {0, 0, 0};
		int subw[] = {0, 0, 0};

		if (fi ->colorFamily == cmYUV)
		{				
			subw[1] = fi->subSamplingW;
			subh[1] = fi->subSamplingH;
			subw[2] = fi->subSamplingW;
			subh[2] = fi->subSamplingH;
		}
		int nbytes = fi->bytesPerSample;
		int nbits = fi->bitsPerSample;
	
		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcpa = vsapi->getReadPtr(srca, plane);
			const uint8_t * srcpb = vsapi->getReadPtr(srcb, plane);
			int stride = vsapi->getStride(srca, plane);
			uint8_t *dstp = vsapi->getWritePtr(dst, plane);
			int ht = vsapi->getFrameHeight(srca, plane);
			int wd = vsapi->getFrameWidth(srca, plane);
			int pitch = stride / nbytes;
			double dy = deltay / (1 << subh[plane]);
			double dx = deltax / (1 << subw[plane]);
			int drpx = dropx >> subw[plane];
			int drpy = dropy >> subh[plane];
	
			if (d->dir == 3)	// down		
			{			//LeftFrame funnels out s wards unmasking RightFrame
				if( fi->sampleType == stInteger && nbytes == 1)
				{
			
					if(2 * bn < d->overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( srcpa + (ht -1-h) * pitch + wd/2, 
												dstp + ( ht - 1 -h) * pitch + wd / 2,
												(int)(h * wd / dy), dy / h);
						for(int h = (int)dy; h < ht; h ++)
							CopyHLinePlanar(srcpa + (ht - 1 - h) * pitch,
											dstp +(ht - 1 - h) * pitch,
											wd);					
					}

					else // if(2*n>=overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( srcpa + (ht -1- drpy - h) * pitch + wd/2, 
												dstp + ( ht - 1 -h) * pitch + wd / 2,
												(int)(h * dx / ht), wd * ht /( h * dx) );											
					}
				}

				else if( fi->sampleType == stInteger && nbytes == 2)
				{
					uint16_t * dp = (uint16_t *) dstp;
					const uint16_t * sp = (const uint16_t *) srcpa;

					if(2 * bn < d->overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( sp + (ht -1-h) * pitch + wd/2, 
												dp + ( ht - 1 -h) * pitch + wd / 2,
												(int)(h * wd / dy), dy / h);
						for(int h = (int)dy; h < ht; h ++)
							CopyHLinePlanar(sp + (ht - 1 - h) * pitch,
											dp +(ht - 1 - h) * pitch,
											wd);					
					}

					else // if(2*n>=overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( sp + (ht -1- drpy - h) * pitch + wd/2, 
												dp + ( ht - 1 -h) * pitch + wd / 2,
												(int)(h * dx / ht), wd * ht /( h * dx) );
											
					}
				}

				else if( fi->sampleType == stFloat)
				{
					float * dp = (float *) dstp;
					const float * sp = (const float *) srcpa;

					if(2 * bn < d->overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( sp + (ht -1-h) * pitch + wd/2, 
												dp + ( ht - 1 -h) * pitch + wd / 2,
												(int)(h * wd / dy), dy / h);
						for(int h = (int)dy; h < ht; h ++)
							CopyHLinePlanar(sp + (ht - 1 - h) * pitch,
											dp +(ht - 1 - h) * pitch,
											wd);					
					}

					else // if(2*n>=overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( sp + (ht -1- drpy - h) * pitch + wd/2, 
												dp + ( ht - 1 -h) * pitch + wd / 2,
												(int)(h * dx / ht), wd * ht /( h * dx) );
											
					}
				}
			}


			else if(d->dir == 1)	// up		
			{
				//A Frame funnels out s wards unmasking B Frame
				if( fi->sampleType == stInteger && nbytes == 1)
				{
					if(2 * bn < d->overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( srcpa + (h) * pitch + wd/2, 
												dstp + (h) * pitch + wd / 2,
												(int)(h * wd / dy), dy / h);
						for(int h = (int)dy; h < ht; h ++)
							CopyHLinePlanar(srcpa + ( h) * pitch,
											dstp + ( h) * pitch,
											wd);					
					}

					else // if(2*n>=overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( srcpa + ( drpy + h) * pitch + wd/2, 
												dstp + ( h) * pitch + wd / 2,
												(int)(h * dx / ht), wd * ht /( h * dx) );
					
					
					}
				}

				else if( fi->sampleType == stInteger && nbytes == 2)
				{
					uint16_t * dp = (uint16_t *) dstp;
					const uint16_t * sp = (const uint16_t *) srcpa;

					if(2 * bn < d->overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( sp + (h) * pitch + wd/2, 
												dp + (h) * pitch + wd / 2,
												(int)(h * wd / dy), dy / h);
						for(int h = (int)dy; h < ht; h ++)
							CopyHLinePlanar(sp + ( h) * pitch,
											dp + ( h) * pitch,
											wd);					
					}

					else // if(2*n>=overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( sp + ( drpy + h) * pitch + wd/2, 
												dp + ( h) * pitch + wd / 2,
												(int)(h * dx / ht), wd * ht /( h * dx) );					
					}
				}

				else if( fi->sampleType == stFloat)
				{
					float * dp = (float *) dstp;
					const float * sp = (const float *) srcpa;

					if(2 * bn < d->overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( sp + (h) * pitch + wd/2, 
												dp + (h) * pitch + wd / 2,
												(int)(h * wd / dy), dy / h);
						for(int h = (int)dy; h < ht; h ++)
							CopyHLinePlanar(sp + ( h) * pitch,
											dp + ( h) * pitch,
											wd);					
					}

					else // if(2*bn>= d->overlap)
					{
						for(int h = 1; h < dy; h ++)

							ResizeHLinePlanar( sp + ( drpy + h) * pitch + wd/2, 
												dp + ( h) * pitch + wd / 2,
										(int)(h * dx / ht), wd * ht /( h * dx) );
					
					
					}
				}
			}

			else if(d->dir == 4)	// left
			{
				if( fi->sampleType == stInteger && nbytes == 1)
				{
					uint8_t * dp = (uint8_t *) dstp;
					const uint8_t * sp = (const uint8_t *) srcpa;
			

					if(2 * bn < d->overlap)
					{
						for(int w = 1; w < dx; w ++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + w,
												dp + pitch * ht / 2 + w,
										(int)(w *  ht / dx), pitch, pitch, dx / w);
						for(int w = (int)dx; w < wd; w ++)

							CopyVLinePlanar(sp + w, dp + w, ht, pitch, pitch);
					
					}				

					else //if(2 * bn >= d->overlap)
					{
						for(int w = 1;w < wd - drpx; w ++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + (drpx + w), 
											dp + pitch * ht / 2 + (w),
								(int)(w * dy / wd), pitch, pitch, wd * ht / (w * dy));

					}
				}

				else if( fi->sampleType == stInteger && nbytes == 2)
				{
					uint16_t * dp = (uint16_t *) dstp;
					const uint16_t * sp = (const uint16_t *) srcpa;
			

					if(2 * bn < d->overlap)
					{
						for(int w = 1; w < dx; w ++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + w,
												dp + pitch * ht / 2 + w,
								(int)(w *  ht / dx), pitch, pitch, dx / w);
						for(int w = (int)dx; w < wd; w ++)

							CopyVLinePlanar(sp + w, dp + w, ht, pitch, pitch);
					
					}				

					else //if(2 * bn >= d->overlap)
					{
						for(int w = 1;w < wd - drpx; w ++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + (drpx + w), 
											dp + pitch * ht / 2 + (w),
								(int)(w * dy / wd), pitch, pitch, wd * ht / (w * dy));

					}
				}

				else if( fi->sampleType == stFloat)
				{
					float * dp = (float *) dstp;
					const float * sp = (const float *) srcpa;
			

					if(2 * bn < d->overlap)
					{
						for(int w = 1; w < dx; w ++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + w,
												dp + pitch * ht / 2 + w,
								(int)(w *  ht / dx), pitch, pitch, dx / w);
						for(int w = (int)dx; w < wd; w ++)

							CopyVLinePlanar(sp + w, dp + w, ht, pitch, pitch);
					
					}				

					else //if(2 * bn >= d->overlap)
					{
						for(int w = 1;w < wd - drpx; w ++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + (drpx + w), 
											dp + pitch * ht / 2 + (w),
								(int)(w * dy / wd), pitch, pitch, wd * ht / (w * dy));

					}
				}

			}		

			else if (d->dir == 2 )	// right
			{
				if( fi->sampleType == stInteger && nbytes == 1)
				{
					uint8_t * dp = (uint8_t *) dstp;
					const uint8_t * sp = (const uint8_t *) srcpa;

					if(2 * bn < d->overlap)
					{
						for(int w=1;w<dx;w++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + (wd - 1 - w),
											dp + pitch * ht / 2 + (wd - 1 - w),
								(int)(w * ht / dx), pitch, pitch, dx / w);
						for(int w = (int)dx; w < wd; w ++)

							CopyVLinePlanar(sp + (wd - 1 - w), dp + (wd - 1 - w), ht, pitch, pitch);
					}

					else // if(2 * bn >= d->overlap)
					{
						for(int w = 1; w < wd - drpx; w ++)

							ResizeVLinePlanar(sp + pitch *ht / 2 + (wd - 1 - (drpx + w)), 
										dp + pitch * ht / 2 + (wd - 1 - w),
								(int)(w * dy / wd), pitch, pitch, wd * ht / (w * dy));

					
					}
				}

				if( fi->sampleType == stInteger && nbytes == 2)
				{
					uint16_t * dp = (uint16_t *) dstp;
					const uint16_t * sp = (const uint16_t *) srcpa;

					if(2 * bn < d->overlap)
					{
						for(int w=1;w<dx;w++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + (wd - 1 - w),
											dp + pitch * ht / 2 + (wd - 1 - w),
								(int)(w * ht / dx), pitch, pitch, dx / w);
						for(int w = (int)dx; w < wd; w ++)

							CopyVLinePlanar(sp + (wd - 1 - w), dp + (wd - 1 - w), ht, pitch, pitch);
					}

					else // if(2 * bn >= d->overlap)
					{
						for(int w = 1; w < wd - drpx; w ++)

							ResizeVLinePlanar(sp + pitch *ht / 2 + (wd - 1 - (drpx + w)), 
										dp + pitch * ht / 2 + (wd - 1 - w),
								(int)(w * dy / wd), pitch, pitch, wd * ht / (w * dy));

					
					}
				}

				if( fi->sampleType == stFloat)
				{
					float * dp = (float *) dstp;
					const float * sp = (const float *) srcpa;

					if(2 * bn < d->overlap)
					{
						for(int w=1;w<dx;w++)

							ResizeVLinePlanar(sp + pitch * ht / 2 + (wd - 1 - w),
											dp + pitch * ht / 2 + (wd - 1 - w),
								(int)(w * ht / dx), pitch, pitch, dx / w);
						for(int w = (int)dx; w < wd; w ++)

							CopyVLinePlanar(sp + (wd - 1 - w), dp + (wd - 1 - w), ht, pitch, pitch);
					}

					else // if(2 * bn >= d->overlap)
					{
						for(int w = 1; w < wd - drpx; w ++)

							ResizeVLinePlanar(sp + pitch *ht / 2 + (wd - 1 - (drpx + w)), 
										dp + pitch * ht / 2 + (wd - 1 - w),
								(int)(w * dy / wd), pitch, pitch, wd * ht / (w * dy));

					
					}
				}

			}
		}
										
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}

//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC funnelFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    FunnelData *d = (FunnelData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC funnelCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    FunnelData d;
    FunnelData *data;
    int err;
	//int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Funnel: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Funnel: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float) vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Funnel: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.dir = int64ToIntS(vsapi->propGetInt(in, "dir", 0, &err));
	if( err)
		d.dir = 1;

	if( d.dir < 1 || d.dir > 4)
	{
		vsapi->setError(out, "Funnel: valid values for dir are 1 to 4 only ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	
	
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (FunnelData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Funnel", funnelInit, funnelGetFrame, funnelFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "funnel", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Funnel", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;", funnelCreate, 0, plugin);
}
*/