/******************************************************************************
Trans.Swing filter plugin function for vapoursynth by V.C.Mohan
4 or 2 or one segments of A Clip  unmasks a stationary B Clip clip by swinging out doors or
4 or 2 or one segments of  B Clip masks  the A clip by swinging in.
 The clips should have identical lengths,  height,  width and same pixel type 
5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation,  version 3 of the License.

    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 

*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap;		// transition length in frames
	bool dir;			// swing dir clockwise or anti
	bool open;	// A swings out or B swings in
	int ndoors;  	// number of symmetric doors
	int corner;	// corner of left top door where hinge is located 
						// 1 top left, 2 top right, 3 bottom right, 4 bottom left
	int vtransStart;
	int vtransEnd;
	int numFrames;
	int xcoord[4];
	int ycoord[4];
	int xoffset[4];
	int yoffset[4];
	int angleSign[4];
	int doorwd;
	int doorht;
	
}SwingData;
//------------------------------------------------------------------
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC swingInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    SwingData *d = (SwingData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

	if( d->ndoors == 1)
	{
		d->doorwd = d->vi.width;
		d->doorht = d->vi.height;
		d->angleSign[0] = d->dir ? 1 : -1;
		d->xoffset[0] = 0;
		d->yoffset[0] = 0;
		if ( d->corner == 1)
		{
			d->xcoord[0] = 0;
			d->ycoord[0] = 0;
		}
		else if ( d->corner == 2)
		{
			d->xcoord[0] = d->doorwd - 1;
			d->ycoord[0] = 0;
		}
		else if ( d->corner == 3)
		{
			d->xcoord[0] = d->doorwd - 1;
			d->ycoord[0] = d->doorht - 1;
		} 

		else if ( d->corner == 4)
		{
			d->xcoord[0] = 0;
			d->ycoord[0] = d->doorht - 1;
		}
	}

	if( d->ndoors == 2)
	{
		d->doorwd = d->vi.width / 2;
		d->doorht = d->vi.height;
		d->xoffset[0] = 0;
		d->yoffset[0] = 0;
		d->xoffset[1] = d->doorwd;
		d->yoffset[1] = 0;
			
		if ( d->corner == 1)
		{
			d->xcoord[0] = 0;
			d->ycoord[0] = 0;
			d->xcoord[1] = d->doorwd - 1;
			d->ycoord[1] = 0;
			d->angleSign[0] = d->dir ? 1 : -1;
		}
		else if ( d->corner == 2)
		{
			d->xcoord[0] = d->doorwd -1;
			d->ycoord[0] = 0;
			d->xcoord[1] = 0;
			d->ycoord[1] = 0;
			d->angleSign[0] = d->dir ? 1 : -1;
		}
		else if ( d->corner == 3)
		{
			d->xcoord[0] = d->doorwd - 1;
			d->ycoord[0] = d->doorht -1;
			d->xcoord[1] = 0;
			d->ycoord[1] = d->doorht -1;
			d->angleSign[0] = d->dir ? -1 : 1;
		} 

		else if ( d->corner == 4)
		{
			d->xcoord[0] = 0;
			d->ycoord[0] = d->doorht - 1;
			d->xcoord[1] =  d->doorwd - 1;
			d->ycoord[1] = d->doorht - 1;
			d->angleSign[0] = d->dir ? -1 : 1;
		}

		d->angleSign[1] =  - d->angleSign[0];
	}

	if( d->ndoors == 4)
	{
		d->doorwd = d->vi.width / 2;
		d->doorht = d->vi.height / 2;
		// direction of rotation of doors
		d->angleSign[0] = d->dir ? 1 : -1;
		d->angleSign[1] =  - d->angleSign[0];
		d->angleSign[2] = - d->angleSign[1];
		d->angleSign[3] = - d->angleSign[2];

		d->xoffset[0] = 0;
		d->yoffset[0] = 0;
		d->xoffset[1] = d->doorwd ;
		d->yoffset[1] = 0;

		d->xoffset[2] = d->doorwd ;
		d->yoffset[2] = d->doorht ;;
		d->xoffset[3] = 0 ;
		d->yoffset[3] = d->doorht ;;

		if ( d->corner == 1)
		{
			d->xcoord[0] = 0;
			d->ycoord[0] = 0;
			d->xcoord[1] = d->doorwd  - 1;
			d->ycoord[1] = 0;
			d->xcoord[2] = d->doorwd  - 1;
			d->ycoord[2] = d->doorht  - 1;
			d->xcoord[3] = 0;
			d->ycoord[3] = d->doorht  - 1;
			
		}
		else if ( d->corner == 2)
		{
			d->xcoord[0] = d->doorwd;
			d->ycoord[0] = 0;
			d->xcoord[1] = 0;
			d->ycoord[1] = 0;
			d->xcoord[2] = 0;
			d->ycoord[2] = d->doorht - 1;
			d->xcoord[3] = d->doorwd;
			d->ycoord[3] = d->doorht - 1;
		}
		else if ( d->corner == 3)
		{
			d->xcoord[0] = d->doorwd - 1;
			d->ycoord[0] = d->doorht  - 1;

			d->xcoord[1] = 0;
			d->ycoord[1] = d->doorht  - 1;

			d->xcoord[2] = 0;
			d->ycoord[2] = 0;

			d->xcoord[3] = d->doorwd - 1;
			d->ycoord[3] = 0;
		} 

		else if ( d->corner == 4)
		{
			d->xcoord[0] = 0;
			d->ycoord[0] = d->doorht  - 1;

			d->xcoord[1] = d->doorwd  - 1;
			d->ycoord[1] = d->doorht  - 1;

			d->xcoord[2] = d->doorwd  - 1;
			d->ycoord[2] = 0;

			d->xcoord[3] = 0;
			d->ycoord[3] = 0;
		} 
		
	}
	if(d->vi.format->colorFamily == cmRGB)
	{
		for( int i = 0; i < d->ndoors; i ++)
		{
			d->ycoord[i] = d->doorht - 1 - d->ycoord[i];
			d->angleSign[i] = - d->angleSign[i];
		}
	}

}

//--------------------------------------------------------------------------------	
static const VSFrameRef *VS_CC swingGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     SwingData *d = (SwingData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		// B frame will be revealed during transition
		
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;	

		VSFrameRef * dst = d->open ? vsapi->copyFrame( srcb, core) : vsapi->copyFrame( srca, core);
		const VSFrameRef *src = d->open ? srca: srcb;
		int degree = ((bn + 1 ) * 90) / (d->overlap + 1) ;
		if (! d->open)
			degree = 90 -  degree; 
		uint8_t * dstp[] = {NULL, NULL, NULL};
		const uint8_t * srcp[] = {NULL, NULL, NULL};
		int pitch[3];

		int subH = fi->colorFamily == cmYUV ? fi->subSamplingH : 0;
		int subW = fi->colorFamily == cmYUV ? fi->subSamplingW : 0;
		int andH = (1 << subH) -1;
		int andW = (1 << subW) -1;

		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			srcp[plane] = vsapi->getReadPtr(src, plane);
			dstp[plane] = vsapi->getWritePtr(dst, plane);
			int stride = vsapi->getStride(srca, plane);
			pitch[plane] = stride / kb;
		}

		int radius = (int)sqrt((float)(d->doorwd * d->doorwd + d->doorht * d->doorht));

		if( fi->sampleType == stInteger && nb == 8)
		{

			for ( int nd = 0; nd < d->ndoors; nd ++)
			{			

				if( fi->numPlanes == 1)
				{
					int offset = d->yoffset[nd] * pitch[0] + d->xoffset[nd];

					vsDiscoRotatePlane(srcp[0] + offset, dstp[0] + offset, radius,
										d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
										d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
										pitch[0], pitch[0], d->angleSign[nd] * degree);
				}
				else
				{
					int offset = d->yoffset[nd] * pitch[0] + d->xoffset[nd];
					int offsetUV = ( d->yoffset[nd] >> subH) * pitch[1] + (d->xoffset[nd] >> subW);

					vsDiscoRotateYUVPlanes(srcp[0] + offset, dstp[0] + offset, radius,
											d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
											d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
											pitch[0], pitch[0], d->angleSign[nd] * degree,
											srcp[1] + offsetUV, srcp[2] + offsetUV, pitch[1], 
											dstp[1] + offsetUV, dstp[2] + offsetUV, pitch[1],
											subH, subW);
				}
			}	// for nd =		
		}	// sampletype integer nb = 8

		else if( fi->sampleType == stInteger && nb > 8)
		{

			uint16_t * dp = (uint16_t *)dstp[0], *dpU = (uint16_t *)dstp[1], *dpV = (uint16_t *) dstp[2];
			const uint16_t * sp = (const uint16_t *)srcp[0], *spU = (const uint16_t *)srcp[1], *spV = (const uint16_t *) srcp[2];

			for ( int nd = 0; nd < d->ndoors; nd ++)
			{			

				if( fi->numPlanes == 1)
				{
					int offset = d->yoffset[nd] * pitch[0] + d->xoffset[nd];

					vsDiscoRotatePlane(sp + offset, dp + offset, radius,
										d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
										d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
										pitch[0], pitch[0], d->angleSign[nd] * degree);
				}
				else
				{
					int offset = d->yoffset[nd] * pitch[0] + d->xoffset[nd];
					int offsetUV = ( d->yoffset[nd] >> subH) * pitch[1] + (d->xoffset[nd] >> subW);

					vsDiscoRotateYUVPlanes(sp + offset, dp + offset, radius,
											d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
											d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
											pitch[0], pitch[0], d->angleSign[nd] * degree,
											spU + offsetUV, spV + offsetUV, pitch[1], 
											dpU + offsetUV, dpV + offsetUV, pitch[1],
											subH, subW);
				}
			}	// for nd =		
		}	// integer nb > 8

		else if( fi->sampleType == stFloat)
		{
			float * dp = (float *)dstp[0], *dpU = (float *)dstp[1], *dpV = (float *) dstp[2];
			const float * sp = (const float *)srcp[0], *spU = (const float *)srcp[1], *spV = (const float *) srcp[2];

			for ( int nd = 0; nd < d->ndoors; nd ++)
			{			

				if( fi->numPlanes == 1)
				{
					int offset = d->yoffset[nd] * pitch[0] + d->xoffset[nd];

					vsDiscoRotatePlane(sp + offset, dp + offset, radius,
										d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
										d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
										pitch[0], pitch[0], d->angleSign[nd] * degree);
				}
				else
				{
					int offset = d->yoffset[nd] * pitch[0] + d->xoffset[nd];
					int offsetUV = ( d->yoffset[nd] >> subH) * pitch[1] + (d->xoffset[nd] >> subW);

					vsDiscoRotateYUVPlanes(sp + offset, dp + offset, radius,
											d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
											d->xcoord[nd], d->ycoord[nd], d->doorht, d->doorwd,
											pitch[0], pitch[0], d->angleSign[nd] * degree,
											spU + offsetUV, spV + offsetUV, pitch[1], 
											dpU + offsetUV, dpV + offsetUV, pitch[1],
											subH, subW);
				}
			}	// for nd =		
		}	// float

		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
	
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC swingFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    SwingData *d = (SwingData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC swingCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    SwingData d;
    SwingData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Swing: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Swing: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Swing: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	temp = !! vsapi->propGetInt(in, "dir", 0, &err);

	if( err)
		d.dir = true;	// clockwise

	else if(  temp < 0 || temp > 1 )
	{
		vsapi->setError(out, "Swing: dir valid value is 0 or 1 for clockwise, anticlockwise rotation");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	else if (temp == 1)

		d.dir = true;
	else
		d.dir = false;

	temp = !! vsapi->propGetInt(in, "open", 0, &err);

	if( err)
		d.open = true;	// open swing out doors

	else if(  temp < 0 || temp > 1 )
	{
		vsapi->setError(out, "Swing: open valid value is 0 or 1 for transition by opening or closing doors");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	else if (temp == 1)

		d.open = true;
	else
		d.open = false;

	d.ndoors =  int64ToIntS(vsapi->propGetInt(in, "ndoors", 0, &err));

	if( err)
		d.ndoors = 1;	// open swing out doors

	else if(  d.ndoors < 1 || d.ndoors > 4  || d.ndoors == 3)
	{
		vsapi->setError(out, "Swing: ndoors valid values are 1 or 2 or 4 only");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	d.corner =  int64ToIntS(vsapi->propGetInt(in, "corner", 0, &err));

	if( err)
		d.corner = 1;	// swing fulcrum at corner 1

	else if(  d.corner < 1 || d.corner > 4 )
	{
		vsapi->setError(out, "Swing: corner valid values are 1 or 2 or 3 or 4 only");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (SwingData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Swing", swingInit, swingGetFrame, swingFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "swing", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Swing", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;open:int:opt;ndoors:int:opt;corner:int:opt;", swingCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------

