
/*
Copyright (C) <2006, 2007,2008,2009>  <V.C.Mohan>

  Auyhor V.C.Mohan

  created/ modified on
   Feb 2006, 2007,2008, 27 Aug 2009

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    http://www.gnu.org/licenses/.
	For details of how to contact author see <http://www.avisynth.org/vcmohan/vcmohan.html> 
*/
/*********************************************************************************************
LineZ contain a set of functions that operate on a single line (vertical or Horizontal)
These are 1 Copy, 2 Flip and magnify linear or as a lens or constant value; Shading
can also be applied. These functions do not check the parameters and assume them to be correct.
Input and out pointers to the middle(except for copy and plain flip functions)
 of window, magnification, window winh( height). pitch for vertical functions  flip, shade are 
some of the parameters needed. 
*/
//*************************************************************************************************
#ifndef vLinez_V_C_MOHAN
#define vLinez_V_C_MOHAN


//----------------------------------------------------------------------------------------------------------------


//			Planar functions
template <typename finc>
inline void CopyHLinePlanar( const finc* srcp, finc* dstp, int winw);
template <typename finc>
inline void FlipHLinePlanar( const finc* srcp, finc* dstp, int winw);
template <typename finc>
inline void ResizeHLinePlanar( const finc* srcp, finc* dstp, int winw,double factor);
template <typename finc>
inline void FlipResizeHLinePlanar( const finc* srcp, finc* dstp, int winw,double factor);
template <typename finc>
inline void FlipLenzHLinePlanar( const finc* srcp, finc* dstp, int winw,double factor);
template <typename finc>
inline void LenzHLinePlanar( const finc* srcp, finc* dstp, int winw,double factor);
template <typename finc>
inline void RollzHLinePlanar( const finc* srcp, finc* dstp, int winw,finc shade);
template <typename finc>
inline void ShadeHLinePlanar( const finc* srcp, finc* dstp, int winw,int shade);
template <typename finc>
void drawHLinePlanar( finc * dp,  int wd, finc val);
//-------------------------------------------------------------------------------------
template <typename finc>
inline void CopyVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch);
template <typename finc>
inline void FlipVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch);
template <typename finc>
inline void ResizeVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,double factor);
template <typename finc>
inline void FlipResizeVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,double factor);
template <typename finc>
inline void FlipLenzVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,double factor);
template <typename finc>
inline void LenzVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,double factor);
template <typename finc>
inline void RollzVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,finc shade);
template <typename finc>
inline void ShadeVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,int shade);
template <typename finc>
void drawVLinePlanar( finc * dp, int pitch, int ht, finc val);


//-----------------------------------------------------------------------------------------------------------------
template <typename finc>
inline void CopyHLinePlanar( const finc* srcp, finc* dstp, int winw)
{
	for(int w=0;w<winw;w++)
		*(dstp+w)=*(srcp+w);
}
template <typename finc>
inline void FlipHLinePlanar( const finc* srcp, finc* dstp, int winw)
{
	for(int w=0; w<winw; w++)
		*(dstp+w)=*(srcp+(winw-1-w));
}
// ResizeHLinePlanar requires the srcp and dstp to point to middle of window winw/2. The factor is input widow width/winw
template <typename finc>
inline void ResizeHLinePlanar( const finc* srcp, finc* dstp, int winw,double factor)
{
						// nearest point interpolation	
	for(int w=0; w<winw/2; w++)
	{
		
		
		*(dstp+w)=*(srcp+(int)(factor*w));
		*(dstp-w)=*(srcp-(int)(factor*w));
		
	}
}
// FlipResizeHLinePlanar requires the srcp and dstp to point to middle of window winw/2. The factor is frame width/winw
template <typename finc>
inline void FlipResizeHLinePlanar( const finc* srcp, finc* dstp, int winw,double factor)
{
						// nearest point interpolation	
	for(int w=0; w<winw/2; w++)
	{
		*(dstp+w)=*(srcp-(int)(factor*w));
		*(dstp-w)=*(srcp+(int)(factor*w));
	}
}
// FlipLenzHLinePlanar requires the srcp and dstp to point to middle of source window and destination windows
// The factor is source window width/destination window width winw
template <typename finc>
inline void FlipLenzHLinePlanar( const finc* srcp, finc* dstp, int winw,double factor)
{
	for(int w=0; w<winw/2; w++)
	{
		
			*(dstp+w)=*(srcp-((int)((factor+(1-factor)*w*2/winw)*w)));
			*(dstp-w)=*(srcp+((int)((factor+(1-factor)*w*2/winw)*w)));
		
	}
}
// LenzHLinePlanar requires the srcp and dstp to point to middle of source window and destination windows
// The factor is source window width/destination window width winw
template <typename finc>
inline void LenzHLinePlanar( const finc* srcp, finc* dstp, int winw,double factor)
{
	for(int w=0; w<winw/2; w++)
	{
		
			*(dstp+w)=*(srcp+((int)((factor+(1-factor)*w*2/winw)*w)));
			*(dstp-w)=*(srcp-((int)((factor+(1-factor)*w*2/winw)*w)));
		
	}
}
// PeelzHLinePlanar requires the srcp and dstp to point to middle of source window and destination windows.
// the destw normally should be touching the srcwindow towards peel direction
//source window should be double that of destination window winw.
//The line is flipped horizontally. Shading can be done optionally
// works for integer formats only
template <typename finc>
inline void RollzHLinePlanar( const finc* srcp, finc* dstp, int winw,finc shade)
{
	int wwd = (shade>=0?winw:0);
	for(int w=0; w<winw/2; w++)
	{
		finc val = ((shade*(wwd-2*w))/winw);
		*(dstp+w)=*(srcp-((int)(w*(1+(double)(2*w/winw)))));
		if ( *(dstp+w) < val)
			*(dstp+w) = val;
		*(dstp-w)=*(srcp+((int)(w*(1+(double)(2*w/winw))))); 
		if ( *(dstp-w) < val)
			*(dstp-w) = val;
		
	}
}
//ShadeHLinePlanar adds shading to a line. Center gets whiter compared to edges.
// the pointers point the center of window.
// works for integer formats only
template <typename finc>
inline void ShadeHLinePlanar( const finc* srcp, finc* dstp, int winw,int shade)
{
	int wwd = (shade>=0?winw:0);
	for(int w=0; w<winw/2; w++)
	{
		
			*(dstp+w)=(*(srcp+w)) | (finc)((shade*(wwd-2*w))/winw);
			*(dstp-w)=(*(srcp-w)) | (finc)((shade*(wwd-2*w))/winw);
		
	}
}

//--------------------------------------------------------------------------------------


template <typename finc>
void drawHLinePlanar( finc * dp,  int wd, finc val)
{
	for (int w = 0; w < wd; w ++)
	{
		dp[w] = val;
	}
}
//***************************************************************************************************************
//Functions for Vertical lines 
// CopyVLinePlanar expects the srcp and dstp to be at the end point of the segment to be copied.
template <typename finc>
inline void CopyVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch)
{
	for(int h=0;h<winh;h++)
		*(dstp+dpitch*h)=*(srcp+spitch*h);
}
//FlipVLinePlanar expects the srcp and dstp to be at the start point of the segment to be copied 
template <typename finc>
inline void FlipVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch)
{
	for(int h=0; h<winh; h++)
		*(dstp+dpitch*h)=*(srcp+spitch*(winh-1-h));
}
// ResizeVLinePlanar requires the srcp and dstp to point to middle of window winh/2. The factor is frame width/winh
template <typename finc>
inline void ResizeVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,double factor)
{
						// nearest point interpolation	
	for(int h=0; h<winh/2; h++)
	{
	
		*(dstp+dpitch*h)=*(srcp+spitch*((int)(factor*h)));
		*(dstp-dpitch*h)=*(srcp-spitch*((int)(factor*h)));
		
	}
}
// FlipResizeVLinePlanar requires the srcp and dstp to point to middle of window winh/2. The factor is frame width/winh
template <typename finc>
inline void FlipResizeVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,double factor)
{
						// nearest point interpolation	
	for(int h=0; h<winh/2; h++)
	{
	
		*(dstp+dpitch*h)=*(srcp-spitch*((int)(factor*h)));
		*(dstp-dpitch*h)=*(srcp+spitch*((int)(factor*h)));
		
	}
}
// FlipLenzVLinePlanar requires the srcp and dstp to point to middle of source window and destination windows
// The factor is source window width/destination window width winh
template <typename finc>
inline void FlipLenzVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,double factor)
{
	for(int h=0; h<winh/2; h++)
	{
		
		*(dstp+dpitch*h)=*(srcp-spitch*((int)((factor+(1-factor)*h*2/winh)*h)));
		*(dstp-dpitch*h)=*(srcp+spitch*((int)((factor+(1-factor)*h*2/winh)*h)));
		
	}
}
// LenzVLinePlanar requires the srcp and dstp to point to middle of source window and destination windows
// The factor is source window width/destination window width winh
template <typename finc>
inline void LenzVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,double factor)
{
	for(int h=0; h<winh/2; h++)
	{
		
		*(dstp+dpitch*h)=*(srcp+spitch*((int)((factor+(1-factor)*h*2/winh)*h)));
		*(dstp-dpitch*h)=*(srcp-spitch*((int)((factor+(1-factor)*h*2/winh)*h)));
		
	}
}
// PeelzVLinePlanar requires the srcp and dstp to point to middle of source window and destination windows.
// the destw normally should be touching the srcwindow towards peel direction
//source window should be double that of destination window winh.
//The line is flipped vertically. Shading can be done optionally  shade= Zero will have no shading
// for integer formats only
template <typename finc>
inline void RollzVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,finc shade)
{
	int wwd = (shade>=0?winh:0);
	for(int h=0; h<winh/2; h++)
	{
		finc val = ((shade*(wwd-2*h))/winh);

		*(dstp+dpitch*h)=*(srcp-spitch*((int)(h*(1+(double)(2*h/winh)))));
		if ( *(dstp+dpitch*h) < val)
			*(dstp+dpitch*h) = val;

		*(dstp-dpitch*h)=*(srcp+spitch*((int)(h*(1+(double)(2*h/winh))))); 
		if ( *(dstp-dpitch*h) < val)
			 *(dstp-dpitch*h) = val;
	}
}
//ShadeVLinePlanar adds shading to a line. Center gets whiter compared to edges.
// the pointers point the center of window. 
// for integer formats only
template <typename finc>
inline void ShadeVLinePlanar( const finc* srcp, finc* dstp, int winh,int spitch, int dpitch,int shade)
{
	int wwd = (shade>=0?winh:0);
	for(int h=0; h<winh/2; h++)
	{
		
			*(dstp+dpitch*h)=(*(srcp+spitch*h)) | (finc)((shade*(wwd-2*h))/winh);
			*(dstp-dpitch*h)=(*(srcp-spitch*h)) | (finc)((shade*(wwd-2*h))/winh);
		
	}
}
//------------------------------------------------------------------------------
template <typename finc>
void drawVLinePlanar( finc * dp, int pitch, int ht, finc val)
{
	for ( int h = 0; h < ht; h ++)
	{
		dp[0] = val;
		dp += pitch;
	}
}
/*
template <typename finc>
inline void ShadePlanarH( const finc* sp, finc* dp, int wd, int shade)
{						
	for(int w = 0;w < wd; w ++)						
	{
		dp[w]
			=(dp[ w] * shade) >> 8;

		dp[(bht - 1 - h) * dpitch + kb * w]
			=(dp[ (bht - 1 - h) * dpitch + kb * w] * shade) >> 8;
									
	}
}
					}
					*/
#endif