/*******************************************************************************
trans dll is a plugin for vapoursynth and provides a number of Scene Transitions.
The input for these are trimmed clips (end of first or A Clip and begining of 
Second or B Clip. Both clips should be of equal length and the transition 
occurs over this length. The clips shpuld have identical color space, width, height
pitch. The output will be one clip of same length as input. This need to be spliced
with the remaining parts of the A Clip (begining to the trim start) and the 
B Clip leftout part after trim.
********************************************************************************/
#include <stdlib.h>
#include "VapourSynth.h"
#include "VSHelper.h"
#define _USE_MATH_DEFINES
#include "math.h"
/******************************************************************************/
//#include <iostream>

#include "vsDiscoRotate.h"
#include "vLinez.h"
#include "vsDisco.h"
#include "vsResizer.h"
#include "UnitSq2Quad_matrix.h"
#include "vsResizeRotateP.h"
#include "vsRipple.h"
#include "vsColorPlaneProcesses.h"
/********************************************************************************/

// Transition Source Files :-

#include "Accord.cpp"

#include "Bubbles.cpp"
#include "Central.cpp"
#include "Crumple.cpp"

#include "Disco.cpp"
#include "Door.cpp"
#include "FlipPage.cpp"
#include "Funnel.cpp"
#include "Paint.cpp"
#include "Roll.cpp"
#include "Push.cpp"
#include "Ripple.cpp"
#include "Scratch.cpp"
#include "Shuffle.cpp"
#include "Slide.cpp"
#include "Sprite.cpp"
#include "Swing.cpp"
#include "Swirl.cpp"
#include "VenetianBlinds.cpp"
#include "Weave.cpp"
#include "Wipe.cpp"

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted

VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin) 
{
    configFunc("in.trans.vcm", "trans", "VapourSynth transition plugin", VAPOURSYNTH_API_VERSION, 1, plugin);

    registerFunc("Accord", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;twin:int:opt;open:int:opt;", accordCreate, 0, plugin);
	registerFunc("Bubbles", "clip:clip;clipb:clip;overlap:float:opt;static:int:opt;", bubblesCreate, 0, plugin);
	registerFunc("Central", "clip:clip;clipb:clip;overlap:float:opt;nturns:int:opt;emerge:int:opt;resize:int:opt;", centralCreate, 0, plugin);
	registerFunc("Crumple", "clip:clip;clipb:clip;overlap:float:opt;crumple:int:opt;emerge:int:opt;", crumpleCreate, 0, plugin);
	registerFunc("Disco", "clip:clip;clipb:clip;overlap:float:opt;emerge:int:opt;rad:int:opt;nturns:int:opt;", discoCreate, 0, plugin);
	registerFunc("Door", "clip:clip;clipb:clip;overlap:float:opt;open:int:opt;vert:int:opt;", doorCreate, 0, plugin);
	registerFunc("FlipPage", "clip:clip;clipb:clip;overlap:float:opt;left:int:opt;", flipPageCreate, 0, plugin);
	registerFunc("Funnel", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;", funnelCreate, 0, plugin);
	registerFunc("Paint", "clip:clip;clipb:clip;overlap:float:opt;style:int:opt;", paintCreate, 0, plugin);
	registerFunc("Roll", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;rollin:int:opt;", rollCreate, 0, plugin);
	registerFunc("Push", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;", pushCreate, 0, plugin);
	registerFunc("Ripple", "clip:clip;clipb:clip;overlap:float:opt;origin:int:opt;amp:int:opt;wlength:int:opt;merge:int:opt:", rippleCreate, 0, plugin);
	registerFunc("Scratch", "clip:clip;clipb:clip;overlap:float:opt;style:int:opt;", scratchCreate, 0, plugin);
	registerFunc("Shuffle", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;", shuffleCreate, 0, plugin);
	registerFunc("Slide", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;slidein:int:opt;", slideCreate, 0, plugin);
	registerFunc("Sprite", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;", spriteCreate, 0, plugin);
	registerFunc("Swing", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;open:int:opt;ndoors:int:opt;corner:int:opt;", swingCreate, 0, plugin);
	registerFunc("Swirl", "clip:clip;clipb:clip;overlap:float:opt;qr:int:opt;dir:int:opt;", swirlCreate, 0, plugin);
	registerFunc("VenitianBlinds", "clip:clip;clipb:clip;overlap:float:opt;style:int:opt;slatwidth:int:opt;open:int:opt;", venitianblindsCreate, 0, plugin);
	registerFunc("Weave", "clip:clip;clipb:clip;overlap:float:opt;style:int:opt;slatwidth:int:opt;", weaveCreate, 0, plugin);
	registerFunc("Wipe", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;", wipeCreate, 0, plugin);
}
