// Plane processing for vapoursynth
#ifndef VSPLANEPROCESS_V_C_MOHAN
#define VSPLANEPROCESS_V_C_MOHAN
//--------------------------------------------------------------------------
// use for greying U V planes
template <typename finc>
void ColorPlane(finc * dp, int dpitch, int width, int height, finc color);
// for single full frame merging
template <typename finc>
void MergePlane(finc * dp,  const finc * spa, const finc * spb,
				int pitch, int width, int height, int mult1, int mult2);

// for single full frame color merging. use for gradual greying or degreying
template <typename finc>
void ColorMergePlane(finc * dp,  const finc * sp, 
				int pitch, int width, int height, 
				int mult1, int mult2, finc color);

// grey RGB frame gradually
template <typename finc>
void GreyMergeRGBFrame(finc * dpr, finc * dpg, finc * dpb,  const finc * spr, const finc * spg,const finc * spb,
				int pitch, int width, int height, int mult1, int mult2);

// grey RGB frame
// grey RGB frame gradually
template <typename finc>
void GreyRGBFrame(finc * dpr, finc * dpg, finc * dpb,  const finc * spr, const finc * spg,const finc * spb,
				int pitch, int width, int height);


//---------------------------------------------------------------------------
template <typename finc>
void ColorPlane(finc * dp, int dpitch, int width, int height, finc color)
{
	for (int h = 0; h < height; h ++)
	{
		for (int w = 0; w < width; w ++)
		{
			dp[w] = color;
		}
		dp += dpitch;
	}
}
//------------------------------------------------------------------------------
// for single full frame merging
template <typename finc>
void MergePlane(finc * dp,  const finc * spa, const finc * spb,
				int pitch, int width, int height, int mult1, int mult2)
{
	for (int h = 0; h < height; h ++)
	{
		for (int w = 0; w < width; w ++)
		{
			dp[w] = ( mult1 * spa[w] + mult2 * spb[w] ) / ( mult1 + mult2);
		}
		dp += pitch;
		spa += pitch;
		spb += pitch;
	}
}

//---------------------------------------------------------------------------
template <typename finc>
void ColorMergePlane(finc * dp,  const finc * sp, 
				int pitch, int width, int height, 
				int mult1, int mult2, finc color)
{
	for (int h = 0; h < height; h ++)
	{
		for (int w = 0; w < width; w ++)
		{
			dp[w] = ( mult1 * sp[w] + mult2 * color ) / ( mult1 + mult2);
		}
		dp += pitch;
		sp += pitch;
		
	}
}
//-----------------------------------------------------------------------------
// grey RGB frame gradually. Assumes G to be Grey value equivalnt to  Y of YUV
template <typename finc>
void GreyMergeRGBFrame(finc * dpr, finc * dpg, finc * dpb,  const finc * spr, const finc * spg,const finc * spb,
				int pitch, int width, int height, int mult1, int mult2)
{
	for (int h = 0; h < height; h ++)
	{
		for (int w = 0; w < width; w ++)
		{
			dpg[w] = spg[w];
			dpr[w] = ( mult1 * spr[w] + mult2 * spg[w] ) / ( mult1 + mult2);
			dpb[w] = ( mult1 * spb[w] + mult2 * spg[w] ) / ( mult1 + mult2);
		}
		dpr += pitch;
		spr += pitch;
		dpg += pitch;
		spg += pitch;
		dpb += pitch;
		spb += pitch;
		
	}
}
//---------------------------------------------------------------------------------
template <typename finc>
void GreyRGBFrame(finc * dpr, finc * dpg, finc * dpb,  const finc * spr, const finc * spg,const finc * spb,
				int pitch, int width, int height)
{
	for (int h = 0; h < height; h ++)
	{
		for (int w = 0; w < width; w ++)
		{
			dpg[w] = spg[w];
			dpr[w] = spg[w];
			dpb[w] = spg[w];
		}
		dpr += pitch;
		spr += pitch;
		dpg += pitch;
		spg += pitch;
		dpb += pitch;
		spb += pitch;
		
	}
}
//---------------------------------------------------------------------------------------------------------


#endif
