/**************************************************************************
vsDisco copies a circular window from source frame to destination frame.
frame pointers, the x,y coordinates of the center of disc in each of
the frames the radius, frame heights and frame widths, pitch sizes and 
 magnification(mag increase from 1 at edge to mag at center, drop magnification(mag increases from 1 at edge to mag at center
 proprtional to dist from center). 
***************************************************************************/
#ifndef vsDisco_V_C_MOHAN
#define vstDisco_V_C_MOHAN
#include "math.h"

//Function
template <typename finc>
bool vsDiscoPlane(const finc* srcp,  finc* dstp, int radius, int srcx, int srcy, int srch, int srcw,
		int dstx, int dsty, int dsth, int dstw, int spitch, int dpitch, double mag=1, bool drop=true, int ringt=0);
template <typename finc>
bool vsDiscoYUVPlanes(const finc* srcp,  finc* dstp, 
						int radius, int srcx, int srcy, const int srch, const int srcw,
						int dstx, int dsty, const int dsth, const int dstw, const int spitch, const int dpitch,
						const finc* srcpU,const finc* srcpV,const int spitchUV,  
						finc* dstpU,  finc* dstpV,const int dpitchUV,
						const int subH, const int subW,
						double mag =1.0 , bool drop = true,int ringt = 0  );
//-------------------------------------------------------------------------------------------------------------------

template <typename finc>
bool vsDiscoPlane(const finc* srcp,  finc* dstp, int radius, int srcx, int srcy, int srch, int srcw,
		int dstx, int dsty, int dsth, int dstw, int spitch, int dpitch, double mag, bool drop, int ringt)
{
	if(srcx+radius<0 || srcx-radius>=srcw || srcy+radius<0 || srcy-radius >=srch 
		|| dstx+radius<0 || dstx-radius>=dstw || dsty+radius<0 || dsty-radius>=dsth || ringt>radius)
		return true;	// no work to be done as either dst or src circle is
						// outside frame area or ring thickness is more than radius
	
	int rt=ringt<=0?radius:ringt;
	int sx=srcx-radius;
	if (sx<1)
		sx=1;
	if(dstx-(srcx-sx)<0)
		sx=sx+srcx-dstx;
	int sy=srcy-radius;
	if (sy<1)
		sy=1;
	if(dsty-(srcy-sy)<0)
		sy=sy+srcy-dsty;
	int ex=srcx+radius;
	if(ex>=srcw)
		ex=srcw-1;
	if(dstx+(ex-srcx)>dstw)
		ex=ex-(dstx-srcx);
	int ey=srcy+radius;
	if(ey>=srch)
		ey=srch-1;
	if(ey+dsty-srcy>dsth)
		ey=ey-(dsty-srcy);
	// we need to process each byte seperately as it is 3 byte per pixel format
	if(mag<=1)
	{

		for(int h=sy;h<ey;h++)
			for(int w=sx; w<ex;w++)
				if((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx)<=radius*radius)
					if((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx)>=(radius-rt)*(radius-rt))

					//	for(int byte=0; byte<3;byte++)
							*(dstp+(h-srcy+dsty)*dpitch+(w-srcx+dstx))
								=*(srcp+(h)*spitch+w);

	
	}

	else if(mag>1)
	{
								// magnification increases towards center
								// search area limited to external square including circle 
		for(int h=sy;h<ey;h++)		
			for(int w=sx; w<ex;w++)
				if((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx)>=(radius-rt)*(radius-rt))
				{
					if((srcx==w && srcy==h)||((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx))==radius*radius)
					{
					
								//if point is on periphery or at center then copy point
						*(dstp+(h-srcy+dsty)*dpitch+(w-srcx+dstx))
							=*(srcp+(h)*spitch+w);
						continue;
					}
					if((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx)<radius*radius )
					{
						double dmag=mag;
						if(drop)

								//when point x,y is within circle area ie x*x +y*y <r*r
								//magnification dmag mag+ (1-mag)* sqrt(hy*hy+wx*wx)/radius
							 dmag=mag + (1.0-mag) * sqrt((double)((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx))) / radius;
						
				
								//if wx is not on y axis wx not equal to zero
								//then value to be taken has x coord of wx/demag, Since this
								//is a real value we need to interpolate between x1 and x2
								//where x1 is integer value of x. X2 is x1+1 if x is +ve
								//and X2 is x1-1 if -ve.
					
					
						double x0=(w-srcx)/dmag;
						int x1=(int)(x0+0.5);
						if(x0<0)
							x1=(int)(x0-0.5);
						
						int addressx1=srcx+x1;
									//If hy is not along x axis ie is non zero
									//then the y coord of value to be taken is h/dmag
									//We need to get nearest y coord							
						double y0=(h-srcy)/dmag;
						int y1=(int)(y0+0.5);
						if(y0<0)
							y1=(int)(y0-0.5);
	
						*(dstp+(h-srcy+dsty)*dpitch+(w-srcx+dstx))
							=*(srcp+(y1+srcy)*spitch+addressx1);
				

					
					}	
				}
		
	}

	return true;
}

//------------------------------------------------------------------------------------
template <typename finc>
bool vsDiscoYUVPlanes(const finc* srcp,  finc* dstp, 
						int radius, int srcx, int srcy, const int srch, const int srcw,
						int dstx, int dsty, const int dsth, const int dstw, const int spitch, const int dpitch,
						const finc* srcpU,const finc* srcpV,const int spitchUV,  
						finc* dstpU,  finc* dstpV,const int dpitchUV,
						const int subH, const int subW,
						double mag  , bool drop ,int ringt   )
{
	int andH = ( 1 << subH ) - 1;
	int andW = ( 1 << subW ) - 1;

	if(srcx+radius<0 || srcx-radius>=srcw || srcy+radius<0 || srcy-radius >=srch 
		|| dstx+radius<0 || dstx-radius>=dstw || dsty+radius<0 || dsty-radius>=dsth || ringt>radius)
		return true;	// no work to be done as either dst or src circle is
						// outside frame area or ring thickness is more than radius
	
	int rt=ringt<=0?radius:ringt;
	int sx=srcx-radius;
	if (sx<1)
		sx=1;
	if(dstx-(srcx-sx)<0)
		sx=sx+srcx-dstx;
	int sy=srcy-radius;
	if (sy<1)
		sy=1;
	if(dsty-(srcy-sy)<0)
		sy=sy+srcy-dsty;
	int ex=srcx+radius;
	if(ex>=srcw)
		ex=srcw-1;
	if(dstx+(ex-srcx)>dstw)
		ex=ex-(dstx-srcx);
	int ey=srcy+radius;
	if(ey>=srch)
		ey=srch-1;
	if(ey+dsty-srcy>dsth)
		ey=ey-(dsty-srcy);
	// we need to process each byte seperately as it is 3 byte per pixel format
	if(mag<=1)
	{
		for(int h=sy;h<ey;h++)
			for(int w=sx; w<ex;w++)
				if((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx)<=radius*radius)
					if((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx)>=(radius-rt)*(radius-rt))
					{
						*(dstp+(h-srcy+dsty)*dpitch+(w-srcx+dstx))							
							=*(srcp+(h)*spitch+w);

						if (  (h & andH) == 0 &&  (w & andW) == 0)
						{

							*(dstpU + ((h-srcy+dsty) >> subH) * dpitchUV + ((w-srcx+dstx) >> subW))							
								= *(srcpU + (h >> subH) * spitchUV + (w >> subW));

							*(dstpV + ((h-srcy+dsty) >> subH) * dpitchUV + ((w-srcx+dstx) >> subW))							
								= *(srcpV + (h >> subH) * spitchUV + (w >> subW));
						}

					}
	}

	else if(mag>1)
	{
								// magnification increases towards center
								// search area limited to external square including circle 
		for(int h=sy;h<ey;h++)		
			for(int w=sx; w<ex;w++)
				if((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx)>=(radius-rt)*(radius-rt))
				{
					if((srcx==w && srcy==h)||((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx))==radius*radius)
					{
					
								//if point is on periphery or at center then copy point
						*(dstp+(h-srcy+dsty)*dpitch+(w-srcx+dstx))
							=*(srcp+(h)*spitch+w);

						*(dstpU + ((h-srcy+dsty) >> subH) * dpitchUV + ((w-srcx+dstx) >> subW))							
								= *(srcpU + (h >> subH) * spitchUV + (w >> subW));

						*(dstpV + ((h-srcy+dsty) >> subH) * dpitchUV + ((w-srcx+dstx) >> subW))							
								= *(srcpV + (h >> subH) * spitchUV + (w >> subW));

						continue;
					}
					if((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx)<radius*radius )
					{
						double dmag=mag;
						if(drop)

								//when point x,y is within circle area ie x*x +y*y <r*r
								//magnification dmag mag+ (1-mag)* sqrt(hy*hy+wx*wx)/radius
								 dmag=mag + (1.0-mag) * sqrt((double)((h-srcy)*(h-srcy)+(w-srcx)*(w-srcx))) / radius;
						
				
								//if wx is not on y axis wx not equal to zero
								//then value to be taken has x coord of wx/demag, Since this
								//is a real value we need to interpolate between x1 and x2
								//where x1 is integer value of x. X2 is x1+1 if x is +ve
								//and X2 is x1-1 if -ve.
					
					
						double x0=(w-srcx)/dmag;
						int x1=(int)(x0+0.5);
						if(x0<0)
							x1=(int)(x0-0.5);
						
						int addressx1=srcx+x1;
									//If hy is not along x axis ie is non zero
									//then the y coord of value to be taken is h/dmag
									//We need to get nearest y coord							
						double y0=(h-srcy)/dmag;
						int y1=(int)(y0+0.5);
						if(y0<0)
							y1=(int)(y0-0.5);
	
						*(dstp+(h-srcy+dsty)*dpitch+(w-srcx+dstx))
							=*(srcp+(y1+srcy)*spitch+addressx1);

						if ( ( h & andH )== 0 &&  (w & andW) == 0)
						{

							*(dstpU + ((h-srcy+dsty) >> subH) * dpitchUV + ((w-srcx+dstx) >> subW))							
								= *(srcpU + (h >> subH) * spitchUV + (w >> subW));

							*(dstpV + ((h-srcy+dsty) >> subH) * dpitchUV + ((w-srcx+dstx) >> subW))							
								= *(srcpV + (h >> subH) * spitchUV + (w >> subW));
						}
					
					}	
				}
		
	}

	return true;
}
#endif
	