/**************************************************************************
VSDiscoRotate copies a circular window from source frame to destination frame
after rotating anticlockwise through an angle degree.
frame pointers, the x,y coordinates of the center of disc in each of
the frames the radius, frame heights and frame widths, pitch sizes and 
the degree
***************************************************************************/
//Function definitions
#ifndef VSDiscoRotate_H_V_C_MOHAN
#define VSDiscoRotate_H_V_C_MOHAN
#include "math.h"
template <typename finc>
bool vsDiscoRotatePlane(const finc *srcp,finc*dstp,
					  int radius,int srcx,int srcy,int srch,int srcw,
					  int dstx,int dsty,int dsth,int dstw,int spitch, int dpitch,int degree);

template <typename finc>
bool vsORingRotatePlane(const finc *srcp,finc*dstp,
					  int radius,int srcx,int srcy,int srch,int srcw,
					  int dstx,int dsty,int dsth,int dstw,int spitch, int dpitch,int degree, int ringt);
template <typename finc>	// color merges two rings
bool vsMergeORingRotatePlane(const finc *srcpa, const finc *srcpb,finc*dstp,
					  int radius,int x0,int y0,int ht,int wd,int pitch,
					  int degree, int ringt, int mult1, int mult2);

template <typename finc>
bool vsDiscoRotateYUVPlanes(const finc *srcp,finc*dstp,
					  int radius,int srcx,int srcy,const int srch,const int srcw,
					  int dstx,int dsty,const int dsth,const int dstw,
					  const int spitch, const int dpitch,int degree,
					  const finc *srcpU,const finc *srcpV,const int spitchUV, 
					  finc*dstpU, finc*dstpV, const int dpitchUV,
					  const int subH, const int subW);

template <typename finc>
bool vsORingRotateYUVPlanes(const finc *srcp,finc*dstp,
					  int radius,int srcx,int srcy,const int srch,const int srcw,
					  int dstx,int dsty,const int dsth,const int dstw,
					  const int spitch, const int dpitch,int degree,int ringt,
					  const finc *srcpU,const finc *srcpV,const int spitchUV, 
					  finc*dstpU, finc*dstpV, const int dpitchUV,
					  const int subH, const int subW);
//===========================================================================================
/**************************************************************************
VSDiscoRotate copies a circular window from source frame to destination frame
after rotating anticlockwise through an angle degree(clockwise if degree is negative).
frame pointers, the x,y coordinates of the center of disc in each of
the frames the radius, frame heights and frame widths, pitch sizes and 
the degree
***************************************************************************/
template <typename finc>
bool vsDiscoRotatePlane(const finc *srcp,finc*dstp,
					  int radius,int srcx,int srcy,int srch,int srcw,
					  int dstx,int dsty,int dsth,int dstw,int spitch, int dpitch,int degree)
{
	
	if(srcx+radius<0 || srcx-radius>=srcw || srcy+radius<0 || srcy-radius >=srch 
		|| dstx+radius<0 || dstx-radius>=dstw || dsty+radius<0 || dsty-radius>=dsth)
		return true;	// no work to be done as either dst or src circle is
						// outside frame area
	//ensure that the part to be copied is within frames
	srcx=srcx & 0xfffffffc;
	srcy= srcy & 0xfffffffc;
	dstx=dstx & 0xfffffffc;
	dsty= dsty & 0xfffffffc;
	radius=radius & 0xfffffffc;
//	int rt=ringt<=0?radius:ringt;
	int sx=srcx-radius;
	if (sx<1)
		sx=1;
	if(dstx-(srcx-sx)<0)
		sx=sx+srcx-dstx;
	int sy=srcy-radius;
	if (sy<1)
		sy=1;
	if(dsty-(srcy-sy)<0)
		sy=sy+srcy-dsty;
	int ex=srcx+radius;
	if(ex>=srcw)
		ex=srcw-1;
	if(dstx+(ex-srcx)>dstw)
		ex=ex-(dstx-srcx);
	int ey=srcy+radius;
	if(ey>=srch)
		ey=srch-1;
	if(ey+dsty-srcy>dsth)
		ey=ey-(dsty-srcy);

	
	degree = degree%360;
		// no rotation
	if (degree<0)
		degree=degree+360;
	double PI=3.1415;
	double alfa = (double)degree*PI/180;
	double sinalfa = sin(alfa);
	double cosalfa = cos(alfa);

	for(int h=sy; h<ey;h++)
		for(int w=sx; w<ex; w++)
		{
			int x=w-srcx;
			int y=h-srcy;
			if(h==srcy && w== srcx)	// Origin gets copied as it is
			{
				*(dstp+(y+dsty)*dpitch+(x+dstx))
					=*(srcp+h*spitch +w);
				continue;
			}
			// for points inside circle only
			if(y*y+x*x<=radius*radius )
			{
				

				if(degree==0)
				{
					*(dstp+(y+dsty)*dpitch+(x+dstx))
					=*(srcp+h*spitch+w);
					continue;
				}

				
				if(degree==180)
				{
					if(-y+srcy<sy || -y+srcy>ey || -x+srcx<sx || -x+srcx>ex)
						continue;
							// x=-x, y=-y
					*(dstp+(y+dsty)*dpitch+(x+dstx))
					=*(srcp+(-y+srcy)*dpitch+(-x+srcx));
					continue;
				}
				if(degree==90)
				{
					if(x+srcy>ey || -y+srcx >ex || x+srcy<sy || -y+srcx<sx )
						continue;
							//x=-y, y=x
					*(dstp+(y+dsty)*dpitch+(x+dstx))
					=*(srcp+(x+srcy)*spitch+(-y+srcx));
					continue;
				}
				if(degree==270)
				{
					if(-x+srcy>ey || y+srcx >ex || -x+srcy<sy || y+srcx<sx )
						continue;
							// x=y, y=-x
					*(dstp+(y+dsty)*dpitch+(x+dstx))
					=*(srcp+(-x+srcy)*dpitch+(y+srcx));
					continue;
				}
		
			
				double newx = x * cosalfa - y * sinalfa;
	
				int newx1=((int)(newx+0.5)) & 0xfffffffe;
				if(newx1+srcx < sx || newx1+srcx>= ex-1)
					continue;			// point outside circle
				
				double newy = x * sinalfa + y * cosalfa;
				
				int newy1=(int)(newy+0.5);
				if(newy1+srcy < sy || newy1+srcy >= ey-1)
					continue;			// point outside circle

					*(dstp+(y+dsty)*dpitch+(x+dstx))
						= *(srcp+(newy1+srcy)*spitch+(newx1+srcx));

				
			
			}
		}

	return true;
}

/*********************************************************************************************/
template <typename finc>
bool vsDiscoRotateYUVPlanes(const finc *srcp,finc*dstp,
					  int radius,int srcx,int srcy,const int srch,const int srcw,
					  int dstx,int dsty,const int dsth,const int dstw,
					  const int spitch, const int dpitch,int degree,
					  const finc *srcpU,const finc *srcpV,const int spitchUV, 
					  finc*dstpU, finc*dstpV, const int dpitchUV,
					  const int subH, const int subW)
{
	// rotates a given circular disc area in the src frame through degree angle in anti
	// clockwise direction.

	if(srcx+radius<0 || srcx-radius>=srcw || srcy+radius<0 || srcy-radius >=srch 
		|| dstx+radius<0 || dstx-radius>=dstw || dsty+radius<0 || dsty-radius>=dsth)
		return true;	// no work to be done as either dst or src circle is
						// outside frame area

	//ensure that the part to be copied is within frames
	srcx=srcx & 0xfffffffc;
	srcy= srcy & 0xfffffffc;
	dstx=dstx & 0xfffffffc;
	dsty= dsty & 0xfffffffc;
	radius=radius & 0xfffffffc;

	int sx=srcx-radius;
	if (sx<1)
		sx=1;
	if(dstx-(srcx-sx)<0)
		sx=sx+srcx-dstx;
	int sy=srcy-radius;
	if (sy<1)
		sy=1;
	if(dsty-(srcy-sy)<0)
		sy=sy+srcy-dsty;
	int ex=srcx+radius;
	if(ex>=srcw)
		ex=srcw-1;
	if(dstx+(ex-srcx)>dstw)
		ex=ex-(dstx-srcx);
	int ey=srcy+radius;
	if(ey>=srch)
		ey=srch-1;
	if(ey+dsty-srcy>dsth)
		ey=ey-(dsty-srcy);

	int andH = (1 << subH) - 1;
	int andW = (1 << subW) - 1;

	degree = degree%360;
		// no rotation
	if (degree<0)
		degree=degree+360;
	double PI=3.1415;
	double alfa = (double)degree*PI/180;
	double sinalfa = sin(alfa);
	double cosalfa = cos(alfa);

	for(int h = sy; h < ey;h ++)

		for(int w = sx; w < ex; w ++)
		{
			int x = w - srcx;
			int y = h - srcy;

			if(h == srcy && w == srcx)	// Origin gets copied as it is
			{
				*(dstp + (y + dsty) * dpitch + (x + dstx))
					=*(srcp + h * spitch + w);
				*(dstpU + ((y + dsty)>> subH) * dpitchUV + ((x + dstx)>> subW))
					=*(srcpU + (h >> subH) * spitchUV + (w >> subW));
				*(dstpV + ((y + dsty)>> subH) * dpitchUV + ((x + dstx)>> subW))
					=*(srcpV + (h >> subH) * spitchUV + (w >> subW));
				
			}
			// for points inside circle only
			else if(y*y+x*x<=radius*radius )
			{
				
			
				double newx = x * cosalfa - y * sinalfa;
	
				int newx1=((int)(newx+0.5)) & 0xfffffffe;

				if(newx1+srcx < sx || newx1+srcx>= ex-1)

					continue;			// point outside circle
				
				double newy = x * sinalfa + y * cosalfa;
				
				int newy1=(int)(newy+0.5);

				if(newy1+srcy < sy || newy1+srcy >= ey-1)

					continue;			// point outside circle

				*(dstp + (y + dsty) * dpitch + (x + dstx))
					=*(srcp + (newy1 + srcy) * spitch + (newx1 + srcx));

				if ( ((y + dsty) & andH) == 0 && ((x + dstx) & andW) == 0)
				{
					*(dstpU + ((y + dsty)>> subH) * dpitchUV + ((x + dstx)>> subW))
						=*(srcpU + ((newy1 + srcy) >> subH) * spitchUV + ((newx1 + srcx) >> subW));

					*(dstpV + ((y + dsty)>> subH) * dpitchUV + ((x + dstx)>> subW))
						=*(srcpV + ((newy1 + srcy) >> subH) * spitchUV + ((newx1 + srcx) >> subW));
				}

			
			}
		}

	return true;
}
//--------------------------------------------------------------------------------------------------------
template <typename finc>
bool vsORingRotatePlane(const finc *srcp,finc*dstp,
					  int radius,int srcx,int srcy,int srch,int srcw,
					  int dstx,int dsty,int dsth,int dstw,int spitch, int dpitch,int degree, int ringt)
{
	
	if(srcx+radius<0 || srcx-radius>=srcw || srcy+radius<0 || srcy-radius >=srch 
		|| dstx+radius<0 || dstx-radius>=dstw || dsty+radius<0 || dsty-radius>=dsth)
		return true;	// no work to be done as either dst or src circle is
						// outside frame area
	//ensure that the part to be copied is within frames
	srcx=srcx & 0xfffffffc;
	srcy= srcy & 0xfffffffc;
	dstx=dstx & 0xfffffffc;
	dsty= dsty & 0xfffffffc;
	radius=radius & 0xfffffffc;
	int rt = ringt - radius;
	if ( ringt <= 0 || ringt >= radius)
		return false;
	int sx=srcx-radius;
	if (sx<1)
		sx=1;
	if(dstx-(srcx-sx)<0)
		sx=sx+srcx-dstx;
	int sy=srcy-radius;
	if (sy<1)
		sy=1;
	if(dsty-(srcy-sy)<0)
		sy=sy+srcy-dsty;
	int ex=srcx+radius;
	if(ex>=srcw)
		ex=srcw-1;
	if(dstx+(ex-srcx)>dstw)
		ex=ex-(dstx-srcx);
	int ey=srcy+radius;
	if(ey>=srch)
		ey=srch-1;
	if(ey+dsty-srcy>dsth)
		ey=ey-(dsty-srcy);

	
	degree = degree%360;
		// no rotation
	if (degree<0)
		degree=degree+360;
	double PI=3.1415;
	double alfa = (double)degree*PI/180;
	double sinalfa = sin(alfa);
	double cosalfa = cos(alfa);
	int radsq = radius * radius;
	int ringsq = (radius - ringt) * (radius - ringt);

	for(int h = sy; h < ey; h ++)
	{
		int y = h - srcy;
		int ysq = y * y;

		for(int w = sx; w < ex; w ++)
		{
			int x = w - srcx;
			// for points inside circle only
			if(ysq + x * x <= radsq && ysq + x * x >= ringsq )
			{			
				float newx = (float)(x * cosalfa - y * sinalfa);
	
				int newx1 = ((int)(newx + 0.5)) & 0xfffffffe;

				if(newx1 + srcx < sx || newx1 + srcx >= ex - 1)
					continue;			// point outside circle
				
				float newy = (float)(x * sinalfa + y * cosalfa);
				
				int newy1 = (int)(newy + 0.5);
				if(newy1 + srcy < sy || newy1 + srcy >= ey - 1)
					continue;			// point outside circle

					*(dstp + (y + dsty) * dpitch + (x + dstx))
						= *(srcp + (newy1 + srcy) * spitch + (newx1 + srcx));

			}
			
		}
	}

	return true;
}
//--------------------------------------------------------------------------------------------

template <typename finc>
bool vsMergeORingRotatePlane(const finc *srcpa, const finc *srcpb,finc * dstp,
					  int radius,int x0,int y0,int ht,int wd,int pitch, 
					  int degree, int ringt, int mult1, int mult2)
{
	
	if(x0 + radius < 0 || x0 - radius >= wd || y0 + radius < 0 || y0 - radius >= ht)
		return true;	// no work to be done as  circle is
						// outside frame area
	//ensure that the part to be copied is within frames
	x0 = x0 & 0xfffffffc;
	y0 =  y0 & 0xfffffffc;	
	radius = radius & 0xfffffffc;

	int rt = ringt - radius;

	if ( ringt <= 0 || ringt >= radius)
		return false;
	int sx = x0 - radius;
	if (sx < 1)
		sx = 1;
	
	int sy = y0- radius;
	if (sy < 1)
		sy = 1;	 
	int ex = x0 + radius;
	if(ex >= wd)
		ex = wd - 1;
	 
	int ey = y0 + radius;
	if(ey >= ht)
		ey = ht - 1;
	
	degree = degree%360;
		
	if (degree < 0)
		degree = degree + 360;

	double PI=3.1415;
	double alfa = (double)degree*PI/180;
	double sinalfa = sin(alfa);
	double cosalfa = cos(alfa);
	int radsq = radius * radius;
	int ringsq = (radius - ringt) * (radius - ringt);

	for(int h = sy; h < ey; h ++)
	{
		int y = h - y0;
		int ysq = y * y;

		for(int w = sx; w < ex; w ++)
		{
			int x = w - x0;
			// for points inside circle only
			if(ysq + x * x <= radsq && ysq + x * x >= ringsq )
			{			
				float newx = (float)(x * cosalfa - y * sinalfa);
	
				int newx1 = ((int)(newx + 0.5)) & 0xfffffffe;

				if(newx1 + x0 < sx || newx1 + x0 >= ex - 1)
					continue;			// point outside circle
				
				float newy = (float)(x * sinalfa + y * cosalfa);
				
				int newy1 = (int)(newy + 0.5);
				if(newy1 + y0 < sy || newy1 + y0 >= ey - 1)
					continue;			// point outside circle

					*(dstp + (y + y0) * pitch + (x + x0))
						= ( (mult1 * (*(srcpa + (newy1 + y0) * pitch + (newx1 + x0)) ) )
						+ (mult2 * (*(srcpb + (newy1 + y0) * pitch + (newx1 + x0)) ) )  )
						/ (mult1 + mult2);

			}
			
		}
	}

	return true;
}
/*********************************************************************************************/
template <typename finc>
bool vsORingRotateYUVPlanes(const finc *srcp,finc*dstp,
					  int radius,int srcx,int srcy,const int srch,const int srcw,
					  int dstx,int dsty,const int dsth,const int dstw,
					  const int spitch, const int dpitch,int degree,int ringt,
					  const finc *srcpU,const finc *srcpV,const int spitchUV, 
					  finc*dstpU, finc*dstpV, const int dpitchUV,
					  const int subH, const int subW)
{
	// rotates a given circular disc area in the src frame through degree angle in anti
	// clockwise direction.
/*********************************************************************************************/
	if(srcx+radius<0 || srcx-radius>=srcw || srcy+radius<0 || srcy-radius >=srch 
		|| dstx+radius<0 || dstx-radius>=dstw || dsty+radius<0 || dsty-radius>=dsth)
		return true;	// no work to be done as either dst or src circle is
						// outside frame area
	if(ringt <= 0 || ringt >= radius)
		return false;
	//ensure that the part to be copied is within frames
	srcx=srcx & 0xfffffffc;
	srcy= srcy & 0xfffffffc;
	dstx=dstx & 0xfffffffc;
	dsty= dsty & 0xfffffffc;
	int radsq = radius *radius;
	int ringsq = (radius - ringt) * (radius - ringt);

	int sx=srcx-radius;
	if (sx<1)
		sx=1;
	if(dstx-(srcx-sx)<0)
		sx=sx+srcx-dstx;
	int sy=srcy-radius;
	if (sy<1)
		sy=1;
	if(dsty-(srcy-sy)<0)
		sy=sy+srcy-dsty;
	int ex=srcx+radius;
	if(ex>=srcw)
		ex=srcw-1;
	if(dstx+(ex-srcx)>dstw)
		ex=ex-(dstx-srcx);
	int ey=srcy+radius;
	if(ey>=srch)
		ey=srch-1;
	if(ey+dsty-srcy>dsth)
		ey=ey-(dsty-srcy);

	int andH = (1 << subH) - 1;
	int andW = (1 << subW) - 1;

	degree = degree%360;
		// no rotation
	if (degree<0)
		degree=degree+360;
	double PI=3.1415;
	double alfa = (double)degree*PI/180;
	double sinalfa = sin(alfa);
	double cosalfa = cos(alfa);

	for(int h = sy; h < ey;h ++)
	{
		int y = h - srcy;
		int ysq = y * y;

		for(int w = sx; w < ex; w ++)
		{
			int x = w - srcx;
			
			if(ysq + x * x <= radius * radius && ysq + x * x >= ringsq )
			{			
				float newx = (float)(x * cosalfa - y * sinalfa);
	
				int newx1 = ((int)(newx + 0.5)) & 0xfffffffe;

				if(newx1 + srcx < sx || newx1 + srcx >= ex - 1)

					continue;			// point outside circle
				
				float newy = (float)(x * sinalfa + y * cosalfa);
				
				int newy1 = (int)(newy + 0.5);

				if(newy1 + srcy < sy || newy1 + srcy >= ey-1)

					continue;			// point outside circle

				*(dstp + (y + dsty) * dpitch + (x + dstx))
					=*(srcp + (newy1 + srcy) * spitch + (newx1 + srcx));

				if ( ((y + dsty) & andH) == 0 && ((x + dstx) & andW) == 0)
				{
					*(dstpU + ((y + dsty)>> subH) * dpitchUV + ((x + dstx)>> subW))
						=*(srcpU + ((newy1 + srcy) >> subH) * spitchUV + ((newx1 + srcx) >> subW));

					*(dstpV + ((y + dsty)>> subH) * dpitchUV + ((x + dstx)>> subW))
						=*(srcpV + ((newy1 + srcy) >> subH) * spitchUV + ((newx1 + srcx) >> subW));
				}			
			}
		}
	}

	return true;
}
#endif		
