/**************************************************************************
vsResizeRotateP copies a circular window from source frame to destination frame
after rotating anticlockwise through an angle degree(clockwise if degree is negative).
frame pointers, the x,y coordinates of the center of disc in each of
the frames the radius, frame heights and frame widths, pitch sizes and 
the degree
***************************************************************************/
#ifndef vsResizeRotateP_V_C_MOHAN
#define vsResizeRotateP_V_C_MOHAN
#include "math.h"


//Function definitions

// for planar formats
template<typename finc>
void ResizeRotatePlane(finc*dstp, int dsth,int dstw, int dpitch,
					  const finc *srcp,int srch,int srcw,int spitch,
					  int degree, double zoom);

	// All planes
template<typename finc>
void ResizeRotatePYUV(finc*dstp, finc*dstpU, finc*dstpV,
					   int dsth,int dstw, int dpitch,int dpitchUV,
					  const finc *srcp, const finc *srcpU,const finc *srcpV,
					  int srch,int srcw,int spitch,int spitchUV,int subH, int subW,
					  int degree, double zoom);

//----------------------------------------------------------------------------------------------------------

/****************************************************/
// for planar formats
template<typename finc>
void ResizeRotatePlane(finc*dstp, int dsth,int dstw, int dpitch,
					  const finc *srcp,int srch,int srcw,int spitch,
					  int degree, double zoom)
{
	// rotates a given circular disc area in the src plane through degree angle 
	//and posts the resized frame on to output frame
	

	degree = degree%360;
	
	if (degree<0)
		degree=degree+360;
	double PI=3.1415;
	double alfa = (double)degree*PI/180;
	double sinalfa = sin(alfa);
	double cosalfa = cos(alfa);

		// Origin gets copied as it is
			
	*(dstp+dsth/2*dpitch+dstw/2)
		
		=*(srcp+srch/2*spitch+srcw/2);

	int radmax = (int)(1.0 / zoom * sqrt((float)(srch * srch + srcw * srcw))/2);

	//radmax &= 0xfffffffc;

	int hmax = (radmax > dsth/2 ? dsth/2 : radmax) ;

	int wmax = (radmax > dstw/2 ? dstw/2 : radmax) ;

	for(int h = - hmax; h < hmax; h++)
	{
		double hsinalfa = (h) * sinalfa; 

		double hcosalfa = (h) * cosalfa;

		for(int w = - wmax; w < wmax; w++)
		{	
			double newx = zoom * ((w) * cosalfa - hsinalfa);

			double newy = zoom * ((w) * sinalfa + hcosalfa);

			int sx = (int)(newx + 0.5);

			int sy = (int)(newy + 0.5);

			if( sx >= srcw/2 || sx <= -srcw/2 || sy >= srch/2 || sy <= -srch/2)				

				continue;	// beyond input frame
			
						
			dstp[(dsth/2+h)*dpitch+(dstw/2 + w)]
				= srcp[(srch/2 + sy)*spitch+(srcw/2 + sx)];
			
		}
	}
		
}

//-----------------------------------------------------------------------------------------------------
template<typename finc>
void ResizeRotatePYUV(finc*dstp, finc*dstpU, finc*dstpV,
					   int dsth,int dstw, int dpitch,int dpitchUV,
					  const finc *srcp, const finc *srcpU,const finc *srcpV,
					  int srch,int srcw,int spitch,int spitchUV,int subH, int subW,
					  int degree, double zoom)
{
	// rotates a given circular disc area in the src frame through degree angle 
	//and posts the resized frame on to output frame. All planar planes rotated
/*********************************************************************************************/
	
	int andH = (1 << subH) -1;

	int	andW = (1 << subW) -1;

	degree = degree%360;
	
	if (degree<0)
		degree=degree+360;
	double PI=3.1415;
	double alfa = (double)degree*PI/180;
	double sinalfa = sin(alfa);
	double cosalfa = cos(alfa);

		// Origin gets copied as it is
			
	*(dstp+dsth/2*dpitch+dstw/2) 
		=*(srcp+srch/2*spitch+srcw/2);

	*(dstpU+ (dsth >> subH)/2*dpitchUV+(dstw >> subW)/2) 
		=*(srcpU+(srch >> subH)/2*spitchUV+(srcw >> subW)/2);
	*(dstpV+(dsth >> subH)/2*dpitchUV+(dstw >> subW)/2) 
		=*(srcpV+(srch >> subH)/2*spitchUV+(srcw >> subW)/2);

	int radmax = (int)(1.0 / zoom * sqrt((float)(srch * srch + srcw * srcw))/2);

	//radmax &= 0xfffffffc;

	int hmax = (radmax > dsth/2 ? dsth/2 : radmax) & 0xfffffffc;

	int wmax = (radmax > dstw/2 ? dstw/2 : radmax) & 0xfffffffc;

	for(int h = - hmax; h < hmax; h++)
	{
		double hsinalfa = (h) * sinalfa;
		double hcosalfa = (h) * cosalfa;

		for(int w = - wmax; w < wmax; w++)
		{	
			double newx = zoom * ((w) * cosalfa - hsinalfa);

			double newy = zoom * ((w) * sinalfa + hcosalfa);

			int sx = (int)(newx + 0.5);
			int sy = (int)(newy + 0.5);

			if( sx >= srcw/2 || sx <= -srcw/2 || sy >= srch/2 || sy <= -srch/2)				

				continue;	// beyond input frame
			
						
			dstp[(dsth/2+h)*dpitch+(dstw/2 + w)]
				= srcp[(srch/2 + sy)*spitch+(srcw/2 + sx)];

			if ( ( h & andH) == 0 && (w & andW) == 0)
			{

				

				dstpU[((dsth/2+h) >> subH) * dpitchUV + ((dstw/2 + w) >> subW)]
						= srcpU[( (srch/2 + sy) >> subH) * spitchUV + ((srcw/2 + sx) >> subW)];


				dstpV[((dsth/2+h) >> subH) * dpitchUV + ((dstw/2 + w) >> subW)]
						= srcpV[( (srch/2 + sy) >> subH) * spitchUV + ((srcw/2 + sx) >> subW)];;
			}
				
			
		}
	}
		
}
#endif		