/***************************************************************************************************
This set of FastResizer functions resizes a source window to the size of the destination window. Nearest point 
method is used for fast work.YUY2 RGB 24 and 
RGB32  formats are supported. The src and dst  buffers must be of adequate sizes. The buffer pointers
point to start of image (left top). The image can also be flipped
***************************************************************************************************/
#ifndef vsResizer_H_V_C_MOHAN
#define vsResizer_H_V_C_MOHAN
// declarations
template <typename finc>
bool HFastResizerPlanar(const finc * srcptr, finc * dstptr,
	int srcht, int srcwd,int dstht, int dstwd, int spitch,
	int dpitch,  bool hflip = false);
template <typename finc>
bool VFastResizerPlanar(const finc * srcptr, finc * dstptr,int srcht, int srcwd,int dstht, int dstwd, int spitch, int dpitch,  bool vflip=false);
		 

/**************************************************************************************************************/
// definitions
//----------------------------------------------------------------------------------
template <typename finc>
bool HFastResizerPlanar(const finc * srcptr, finc * dstptr,
	int srcht, int srcwd,int dstht, int dstwd, int spitch,
	int dpitch,  bool hflip)
{
	if(!(srcht==dstht))
		return false; // these must be equal as only width is being resized
	dstwd &= 0xfffffffe;
	srcwd &= 0xfffffffe;
	
	if (dstwd<2 || srcwd<2)
		return true; // no resizing or flip is necessasary
	
					//calculate magnification/shrink factor
	double hormag = (double)(srcwd)/(double)(dstwd); // srcwd-1 intervals to become dstwd-1 intervals
	int n;
	
	if(hflip)
	{
		
		for(int w=1; w<dstwd; w++) 
										
		{
						// pixel first byte position
			n= (int)(w*hormag+0.5);	
								// calculate source offset for dst offset of w
			for (int h=0;h< srcht; h++)
			{
				*(dstptr+h*dpitch+w)=*(srcptr+h*spitch+(srcwd-n));
						

			}
					
		}
		
			return true;
	}

		
			//The leftmost pixel and Rightmost pixels will be left and right most pixels of output
			// this could not be included in main loop due to access violations
		
	for (int w = 0; w < dstwd; w++) // we will take horizontal symmetry advantage Process byte by byte										
	{
		// pixel first byte position
		n = (int)(w * hormag);
		// calculate source offset for dst offset of w
		for (int h = 0; h < srcht; h++)
		{
			*(dstptr + h * dpitch + w) = *(srcptr + h * spitch + n);


		}

	}
	return true;

}
	

template <typename finc>
bool VFastResizerPlanar(const finc * srcptr, finc * dstptr,
	int srcht, int srcwd,int dstht, int dstwd, int spitch,
	int dpitch,  bool vflip)
{
	if(!(srcwd==dstwd))
		return false; // these must be equal as only height is being resized

	dstht &= 0xfffffffe;
	srcht &= 0xfffffffe;

	if (dstht<2 ||srcht<2)
		return true;	// not possible to resize or flip
	//calculate magnification/shrinkage factor
	double vermag = (double)(srcht)/(double)(dstht);
	int n;
	if(vflip)		
	{

		for (int h=1;h< dstht; h++) 
		{
			n=(int)(h*vermag+0.5);
			for(int w=0; w<dstwd; w++)	
			{
				// sample at 3.4 = sample at 3 
				*(dstptr+w+h*dpitch)=*(srcptr+(srcht-n)*spitch+w);
			}
		}
		return true;
	}

		// no flip
		for (int h=0;h< dstht; h++) 
		{
			n=(int)(h*vermag);
			for(int w=0; w<dstwd; w++)
			{
				// sample at 3.4 = sample at 3 
				*(dstptr+w+h*dpitch)=*(srcptr+n*spitch+w);
			}
		}
	
				 
	return true;
}
#endif
