// Ripple functions
#ifndef VSRIPPLES_H_V_C_MOHAN
#define VSRIPPLES_H_V_C_MOHAN
template <typename finc>
		// for Plane for all formats
	void CreateRipple( finc * wp, int wpitch, const finc * sp, int spitch,
						int xcoord, int ycoord, int irad, int erad, int nn, 
						int bwd, int bht, int amp, int lambda,  float * sintbl);

			// for Planar format

	template <typename finc>
		// for Planar format
	void CreateRipple( finc * wp, int wpitch, finc * wpU,  finc * wpV, int wpitchUV,
						const finc * sp, int spitch, const finc * spU, const finc * spV, int spitchUV,
						int xcoord, int ycoord, int irad, int erad, int nn, 
						int bwd, int bht, int subh, int subw, int amp, int lambda,  float * sintbl);

	template <typename finc>
		// for Planar YUV format
	void ColorGrayRipple( finc * wp, int wpitch, finc * wpU,  finc * wpV, int wpitchUV,
						const finc * sp, int spitch, const finc * spU, const finc * spV, int spitchUV,
						int xcoord, int ycoord, int irad, int erad, int nn,  int mult1, int mult2, finc color, 
						int bwd, int bht, int subh, int subw, int amp, int lambda,  float * sintbl);

	template <typename finc>
		// for Planar RGB format
	void ColorGrayRipple( finc * wpR,  finc * wpG,  finc * wpB, int wpitch,
						const finc * spR, const finc * spG, const finc * spB, int spitch,
						int xcoord, int ycoord, int irad, int erad, int nn,  int mult1, int mult2,  
						int bwd, int bht, int amp, int lambda,  float * sintbl);
	template <typename finc>
		// for Y Plane
	void MergeGrayRipple( finc * wp, int wpitch,
						const finc * spa, int sapitch,const finc * spb, int sbpitch,
						int xcoord, int ycoord, int irad, int erad, int nn,  int mult1, int mult2,  
						int bwd, int bht, int amp, int lambda,  float * sintbl);

	void copyRipples(unsigned char *dp, const int dpitch,
							const unsigned char *sp, const int spitch,
							const int x1, const int y1, const int x2, const int y2, const int rad, 
							const int bht, const int bwd, const int kb);
	void copyRipplesPlanar( unsigned char * dp, int dpitch, unsigned char * dpU,  unsigned char * dpV, int dpitchUV,
							const unsigned char * sp, int spitch, const unsigned char * spU, const unsigned char * spV, int spitchUV,
							const int x1, const int y1, const int x2, const int y2, const int rad, 
							const int bht, const int bwd, const int subH, const int subW);
//--------------------------------------------------------------------------------------------------------------------
template <typename finc>
		// for Planar format
	void CreateRipple( finc * wp, int wpitch, finc * wpU,  finc * wpV, int wpitchUV,
						const finc * sp, int spitch, const finc * spU, const finc * spV, int spitchUV,
						int xcoord, int ycoord, int irad, int erad, int nn, 
						int bwd, int bht, int subh, int subw, int amp, int lambda,  float * sintbl)
	{
		int  bhtUV = bht >> subh;
		int  bwdUV = bwd >> subw;

		int iradsq = irad * irad;
		int eradsq = erad * erad;

		int andh = (1 << subh ) - 1;
		int andw = (1 << subw ) - 1;

		for (int h = ycoord - erad; h <= ycoord + erad; h ++)
		{
			if (h < 0 ) continue;
			if ( h >= bht) break;
			int hsq = (h - ycoord) * (h - ycoord);

			for( int w = (xcoord - erad ); w <= (xcoord + erad) ; w ++)
			{
				if ( w < 0) continue;
				if (w >= bwd ) break;

				int sqr = hsq + (w - xcoord ) * (w - xcoord);
				if ( sqr < iradsq  || sqr > eradsq) continue;
				

				int rad = (int)sqrt((float)sqr);	// radius of circle					
				int rdisp = (int)(sintbl[(rad + nn) % lambda] * amp) ;	// amplitude of sine wave at this radius
				
				if (rad == 0)
				{
					rad = 1;
				}
				
				int	ydisp =  (rdisp * (h - ycoord) ) / rad;	// how much we move in vertical direction				
				int	xdisp =  (rdisp * (w - xcoord) ) / rad;	// how much we move in horizontal direction
				

				if ( h + ydisp < 0 || h + ydisp >= bht || w + xdisp < 0 || w + xdisp >= bwd) continue;

				wp[h * wpitch + w] = sp[ (h + ydisp) * wpitch + w + xdisp];

				if ( (h & andh ) == 0 && (w & andw) == 0)
				{
					wpU[ (h >> subh) * wpitchUV + (w >> subw) ]

					= spU[( (h + ydisp) >> subh) * spitchUV + ( (w + xdisp) >> subw )] ;

					wpV[ (h >> subh) * wpitchUV + (w >> subw) ]

					= spV[( (h + ydisp) >> subh) * spitchUV + ( (w + xdisp) >> subw )];
				}
			}
		} 
	}
//--------------------------------------------------------------------------------------------------
	template <typename finc>
		// for Planar YUV format
	void ColorGrayRipple( finc * wp, int wpitch, finc * wpU,  finc * wpV, int wpitchUV,
						const finc * sp, int spitch, const finc * spU, const finc * spV, int spitchUV,
						int xcoord, int ycoord, int irad, int erad, int nn, int mult1, int mult2, finc color, 
						int bwd, int bht, int subh, int subw, int amp, int lambda,  float * sintbl)
	{
		
		int iradsq = irad * irad;
		int eradsq = erad * erad;

		int andh = (1 << subh ) - 1;
		int andw = (1 << subw ) - 1;

		for (int h = ycoord - erad; h <= ycoord + erad; h ++)
		{
			if (h < 0 ) continue;
			if ( h >= bht) break;
			int hsq = (h - ycoord) * (h - ycoord);

			for( int w = (xcoord - erad ); w <= (xcoord + erad) ; w ++)
			{
				if ( w < 0) continue;
				if (w >= bwd ) break;

				int sqr = hsq + (w - xcoord ) * (w - xcoord);
				if ( sqr < iradsq  || sqr > eradsq) continue;

				int rad = (int)sqrt((float)sqr);	// radius of circle					
				int rdisp = (int)(sintbl[(rad + nn) % lambda] * amp) ;	// amplitude of sine wave at this radius
				
				if (rad == 0)
				{
					rad = 1;
				}
				
				int	ydisp =  (rdisp * (h - ycoord) ) / rad;	// how much we move in vertical direction				
				int	xdisp =  (rdisp * (w - xcoord) ) / rad;	// how much we move in horizontal direction

				if ( h + ydisp < 0 || h + ydisp >= bht ||  w + xdisp < 0 || w + xdisp >= bwd)
				{
					wp[h * wpitch + w] = sp[h * spitch + w];

					if ( (h & andh ) == 0 && (w & andw) == 0)
					{
						wpU[ (h >> subh) * wpitchUV + (w >> subw) ]
							= ( color * mult1 + mult2 * spU[ (h >> subh) * spitchUV + (w >> subw) ]) / (mult1 + mult2);
						wpV[ (h >> subh) * wpitchUV + (w >> subw) ]
							= ( color * mult1 + mult2 * spV[ (h >> subh) * spitchUV + (w >> subw) ]) / (mult1 + mult2);
					}
				}
				else
				{

					wp[h * wpitch + w] = sp[ (h + ydisp) * spitch + w + xdisp];

					if ( (h & andh ) == 0 && (w & andw) == 0)
					{
						wpU[ (h >> subh) * wpitchUV + (w >> subw) ]

						= ( color * mult1 + mult2 * spU[( (h + ydisp) >> subh) * spitchUV + ( (w + xdisp) >> subw )] ) / (mult1 + mult2);

						wpV[ (h >> subh) * wpitchUV + (w >> subw) ]

						= ( color * mult1 + mult2 * spV[( (h + ydisp) >> subh) * spitchUV + ( (w + xdisp) >> subw )] ) / (mult1 + mult2);
					}
				}
			}
		} 
	}
//-------------------------------------------------------------------------------------------
template <typename finc>
		// for Planar RGB format
	void ColorGrayRipple( finc * wpR,  finc * wpG,  finc * wpB, int wpitch,
						const finc * spR, const finc * spG, const finc * spB, int spitch,
						int xcoord, int ycoord, int irad, int erad, int nn,  int mult1, int mult2,  
						int bwd, int bht, int amp, int lambda,  float * sintbl)
	{
		

		int iradsq = irad * irad;
		int eradsq = erad * erad;

		for (int h = ycoord - erad; h <= ycoord + erad; h ++)
		{
			if (h < 0 ) continue;
			if ( h >= bht) break;
			int hsq = (h - ycoord) * (h - ycoord);

			for( int w = (xcoord - erad ); w <= (xcoord + erad) ; w ++)
			{
				if ( w < 0) continue;
				if (w >= bwd ) break;

				int sqr = hsq + (w - xcoord ) * (w - xcoord);
				if ( sqr < iradsq || sqr > eradsq ) continue;
				
				int rad = (int)sqrt((float)sqr);		// radius of circle					
				int rdisp = (int)(sintbl[(rad + nn) % lambda] * amp) ;	// amplitude of sine wave at this radius
				
				if (rad == 0)
				{
					rad = 1;
				}
				
				int	ydisp =  (rdisp * (h - ycoord) ) / rad;	// how much we move in vertical direction				
				int	xdisp =  (rdisp * (w - xcoord) ) / rad;	// how much we move in horizontal direction

				if ( h + ydisp < 0 || h + ydisp >= bht) continue;

				if( w + xdisp < 0 || w + xdisp >= bwd) continue;

				wpG[h * wpitch + w] = spG[ (h + ydisp) * wpitch + w + xdisp];
				wpR[h * wpitch + w] = ( mult1 * wpG[h * wpitch + w] + mult2 * spR[ (h + ydisp) * wpitch + w + xdisp]) / (mult1 + mult2);
				wpB[h * wpitch + w] = (mult1 * wpG[h * wpitch + w] + mult2 * spB[ (h + ydisp) * wpitch + w + xdisp]) / (mult1 + mult2);
			}
		}
	}


//--------------------------------------------------------------------------------------------------
	// for single Plane
	template <typename finc>
	void CreateRipple( finc * wp, int wpitch, const finc * sp, int spitch,
						int xcoord, int ycoord,int irad,  int erad,int nn,  
						int bwd, int bht, int amp, int lambda, float * sintbl)
	{
		
		int iradsq = irad * irad;
		int eradsq = erad * erad;

		for(int h =	0; h < erad; h ++)
		{		
			for(int w = 0; w < erad; w ++)		
			{
				int sqr = h * h + w * w;
				if (sqr < iradsq ) continue;
				if ( sqr > eradsq) break;					
				int rad2x = (int)sqrt((float)sqr);		// radius of circle				
				if(rad2x == 0)					
				rad2x = 1;					// prevent div by zero
				
				int rdisp = (rad2x + nn) % lambda;	// position of wave at this point on this frame
				
				int disp1 = (int)(( sintbl[ rdisp] * amp * h) / rad2x);	// how much we move in vertical direction 
				
				int disp2 = (int)(( sintbl[ rdisp] * amp * w) / rad2x);	// how much we move in horizontal direction 

			
						// use 4 fold symmetry to reduce sqrt computations
						//	(x,y), (x, -y), (-x,y), (-x, -y),

				if(ycoord +( h ) < bht && ycoord +( h + disp1) < bht   &&  ycoord +( h + disp1) >= 0 )
				{							
					if(  xcoord + (w ) < bwd  && xcoord + (w + disp2) < bwd  && xcoord + (w + disp2) >= 0)
					{							
						wp [(ycoord + h) * wpitch + (xcoord + w) ]
							= sp [(ycoord + (h + disp1)) * spitch + (xcoord + (w + disp2)) ];
					}


					if(xcoord - (w ) >= 0 &&  xcoord - (w + disp2) >= 0 && xcoord - (w + disp2) < bwd)
					{
						wp [(ycoord + h) * wpitch + (xcoord - w) ]
							= sp [(ycoord + (h + disp1)) * spitch + (xcoord - (w + disp2)) ];
							
					}							
				}

				if(ycoord - ( h ) >= 0 && ycoord - ( h + disp1) >= 0 && ycoord - ( h + disp1) < bht) 
				{
					if(  xcoord + (w ) < bwd  && xcoord + (w + disp2) < bwd  && xcoord + (w + disp2) >= 0)
					{					
						wp [(ycoord - h) * wpitch + (xcoord + w) ]
							= sp [(ycoord - (h + disp1)) * spitch + (xcoord + (w + disp2)) ];
					}
								
					if(xcoord - (w ) >= 0 &&  xcoord - (w + disp2) >= 0 && xcoord - (w + disp2) < bwd)
					{							
						wp [(ycoord - h) * wpitch + (xcoord - w) ]
							= sp [(ycoord - (h + disp1)) * spitch + (xcoord - (w + disp2)) ];
					}
				}
			}
		}

	}
//-----------------------------------------------------------------------------------------------------------------------
	template <typename finc>
		// for Y Plane. On wp  ripples  merged. 
	void MergeGrayRipple( finc * wp, int wpitch,
						const finc * spa, int sapitch,const finc * spb, int sbpitch,
						int xcoord, int ycoord, int irad, int erad, int nn,  int mult1, int mult2,  
						int bwd, int bht, int amp, int lambda,  float * sintbl)
	{
		int iradsq = irad * irad;
		int eradsq = erad * erad;

		for (int h = ycoord - erad; h <= ycoord + erad; h ++)
		{
			if (h < 0 ) continue;
			if ( h >= bht) break;
			int hsq = (h - ycoord) * (h - ycoord);

			for( int w = (xcoord - erad ); w <= (xcoord + erad) ; w ++)
			{
				if ( w < 0) continue;
				if (w >= bwd ) break;

				int sqr = hsq + (w - xcoord ) * (w - xcoord);
				if ( sqr < iradsq || sqr > eradsq ) continue;
				
				int rad = (int)sqrt((float)sqr);		// radius of circle					
				int rdisp = (int)(sintbl[(rad + nn) % lambda] * amp) ;	// amplitude of sine wave at this radius
				
				if (rad == 0)
				{
					rad = 1;
				}
				
				int	ydisp =  (rdisp * (h - ycoord) ) / rad;	// how much we move in vertical direction				
				int	xdisp =  (rdisp * (w - xcoord) ) / rad;	// how much we move in horizontal direction

				if ( h + ydisp < 0 || h + ydisp >= bht) continue;

				if( w + xdisp < 0 || w + xdisp >= bwd) continue;

				wp[h * wpitch + w] = ( mult1 * spa[ (h + ydisp) * sapitch + w + xdisp] + mult2 * spb[ (h + ydisp) * sbpitch + w + xdisp]) / (mult1 + mult2 );
			}
		}
	}
	


//------------------------------------------------------------------------------------------------------------------------

	void copyRipplesPlanar( unsigned char * dp, int dpitch, unsigned char * dpU,  unsigned char * dpV, const int dpitchUV,
							const unsigned char * sp, int spitch, const unsigned char * spU, const unsigned char * spV, const int spitchUV,
							const int x1, const int y1, const int x2, const int y2, const int rad, 
							const int bht, const int bwd, const int subH, const int subW)
{
	// copies two ripples 
	int radsq = rad * rad;
	for( int h = 0; h < bht; h ++)
	{
		if( (h < y1 - rad || h > y1 + rad) && (h < y2 - rad || h > y2 + rad)) continue;

		int h1sq = ( y1 - h ) * (y1 - h);
		int h2sq = (y2 - h) * (y2 - h);

		if ( h1sq > radsq && h2sq > radsq) continue;
		
		for (int w = 0; w < bwd; w++)
		{
			if( (w < x1 - rad || w > x1 + rad) && (w < x2 - rad || w > x2 + rad)) continue;

		//	int w1sq = (x1 - w) * (x1 - w);
		//	int w2sq = (x2 - w) * (x2 - w);

			if((x1 - w) * (x1 - w) + h1sq > radsq && (x2 - w) * (x2 - w) + h2sq > radsq) continue;

			dp[ h * dpitch + w] = sp[ h * dpitch + w];

			dpU[( h >> subH) * dpitchUV + (w >> subW)] = spU[( h >> subH) * spitchUV + (w >> subW)];
			dpV[( h >> subH) * dpitchUV + (w >> subW)] = spV[( h >> subH) * spitchUV + (w >> subW)];
		}
	}

}
//------------------------------------------------------------------------------------------------------
void copyRipples(unsigned char *dp, const int dpitch,
							const unsigned char *sp, const int spitch,
							const int x1, const int y1, const int x2, const int y2, const int rad, 
							const int bht, const int bwd, const int kb)
{
	// copies two ripples 
	int radsq = rad * rad;
	for( int h = 0; h < bht; h ++)
	{
		if( (h < y1 - rad || h > y1 + rad) && (h < y2 - rad || h > y2 + rad)) continue;

		int h1sq = ( y1 - h ) * (y1 - h);
		int h2sq = (y2 - h) * (y2 - h);

		if ( h1sq > radsq && h2sq > radsq) continue;
		
		for (int w = 0; w < bwd; w++)
		{
			if( (w < x1 - rad || w > x1 + rad) && (w < x2 - rad || w > x2 + rad)) continue;

		//	int w1sq = (x1 - w) * (x1 - w);
		//	int w2sq = (x2 - w) * (x2 - w);

			if((x1 - w) * (x1 - w) + h1sq > radsq && (x2 - w) * (x2 - w) + h2sq > radsq) continue;

			for( int k = 0; k < kb; k ++)
			{

				dp[ h * dpitch + w * kb + k] = sp[ h * dpitch + w * kb + k];
			}

			
		}
	}
}
#endif
